from __future__ import print_function

from .redis_std import StandardEnv
from redis.cluster import ClusterNode
import redis
import time
from RLTest.utils import Colors


class ClusterEnv(object):
    def __init__(self, **kwargs):
        self.shards = []
        self.envIsUp = False
        self.envIsHealthy = False
        self.modulePath = kwargs['modulePath']
        self.moduleArgs = kwargs['moduleArgs']
        self.password = kwargs['password']
        self.shardsCount = kwargs.pop('shardsCount')
        useSlaves = kwargs.get('useSlaves', False)
        self.useTLS = kwargs['useTLS']
        self.decodeResponses = kwargs.get('decodeResponses', False)
        self.tlsPassphrase = kwargs.get('tlsPassphrase', None)
        self.protocol = kwargs.get('protocol', 2)
        self.terminateRetries = kwargs.get('terminateRetries', None)
        self.terminateRetrySecs = kwargs.get('terminateRetrySecs', None)
        startPort = kwargs.pop('port', 10000)
        totalRedises = self.shardsCount * (2 if useSlaves else 1)
        randomizePorts = kwargs.pop('randomizePorts', False)
        for i in range(0, totalRedises, (2 if useSlaves else 1)):
            port = 0 if randomizePorts else startPort
            shard = StandardEnv(port=port, serverId=(i + 1),
                                clusterEnabled=True, **kwargs)
            self.shards.append(shard)
            startPort += 2

    def printEnvData(self, prefix=''):
        print(Colors.Yellow(prefix + 'info:'))
        print(Colors.Yellow(prefix + '\tshards count:%d' % len(self.shards)))
        if self.modulePath:
            print(Colors.Yellow(prefix + '\tzip module path:%s' % self.modulePath))
        if self.moduleArgs:
            print(Colors.Yellow(prefix + '\tmodule args:%s' % self.moduleArgs))
        for i, shard in enumerate(self.shards):
            print(Colors.Yellow(prefix + 'shard: %d' % (i + 1)))
            shard.printEnvData(prefix + '\t')

    def getInformationBeforeDispose(self):
        return [shard.getInformationBeforeDispose() for shard in self.shards]

    def getInformationAfterDispose(self):
        return [shard.getInformationAfterDispose() for shard in self.shards]  

    def waitCluster(self, timeout_sec=40):
        st = time.time()
        ok = 0

        while st + timeout_sec > time.time():
            ok = 0
            for shard in self.shards:
                con = shard.getConnection()
                try:
                    status = con.execute_command('CLUSTER', 'INFO')
                except Exception as e:
                    print('got error on cluster info, will try again, %s' % str(e))
                    continue
                if 'cluster_state:ok' in str(status):
                    ok += 1
            if ok == len(self.shards):
                for shard in self.shards:
                    try:
                        shard.getConnection().execute_command('FT.CLUSTERREFRESH')
                    except Exception:
                        pass
                    try:
                        shard.getConnection().execute_command('SEARCH.CLUSTERREFRESH')
                    except Exception:
                        pass
                return

            time.sleep(0.1)
        raise RuntimeError(
            "Cluster OK wait loop timed out after %s seconds" % timeout_sec)

    def startEnv(self, masters=True, slaves=True):
        if self.envIsUp == True:
            return  # env is already up
        try:
            for shard in self.shards:
                shard.startEnv(masters, slaves)
        except Exception:
            for shard in self.shards:
                shard.stopEnv()
            raise

        slots_per_node = int(16384 / len(self.shards)) + 1
        for i, shard in enumerate(self.shards):
            con = shard.getConnection()
            for s in self.shards:
                con.execute_command('CLUSTER', 'MEET',
                                    '127.0.0.1', s.getMasterPort())

            start_slot = i * slots_per_node
            end_slot = start_slot + slots_per_node
            if end_slot > 16384:
                end_slot = 16384

            try:
                con.execute_command('CLUSTER', 'ADDSLOTS', *(str(x)
                                    for x in range(start_slot, end_slot)))
            except Exception:
                pass

        self.waitCluster()
        self.envIsUp = True
        self.envIsHealthy = True

    def stopEnv(self, masters=True, slaves=True):
        self.envIsUp = False
        self.envIsHealthy = False
        for shard in self.shards:
            shard.stopEnv(masters, slaves)
            self.envIsUp = self.envIsUp or shard.envIsUp
            self.envIsHealthy = self.envIsHealthy and shard.envIsUp

    def getConnection(self, shardId=1):
        return self.shards[shardId - 1].getConnection()

    def getClusterConnection(self):
        statupNode = [ClusterNode(a['host'], a['port']) for a in self.getMasterNodesList()]
        if self.useTLS:
            return redis.RedisCluster(
                ssl=True,
                ssl_keyfile=self.shards[0].getTLSKeyFile(),
                ssl_certfile=self.shards[0].getTLSCertFile(),
                ssl_cert_reqs=None,
                ssl_ca_certs=self.shards[0].getTLSCACertFile(),
                ssl_password=self.tlsPassphrase,
                password=self.password,
                startup_nodes=statupNode,
                decode_responses=self.decodeResponses,
                protocol=self.protocol,
                terminateRetries=self.terminateRetries, terminateRetrySecs=self.terminateRetrySecs
            )
        else:
            return redis.RedisCluster(
                startup_nodes=statupNode,
                decode_responses=self.decodeResponses, password=self.password,
                protocol=self.protocol,
                terminateRetries=self.terminateRetries, terminateRetrySecs=self.terminateRetrySecs)

    def getSlaveConnection(self):
        raise Exception('unsupported')

    # List of nodes that initial bootstrapping can be done from
    def getMasterNodesList(self):
        full_master_list = []
        for shard in self.shards:
            node_info_list = shard.getMasterNodesList()
            full_master_list.append(node_info_list[0])
        return full_master_list

    # List containing a connection for each of the master nodes
    def getOSSMasterNodesConnectionList(self):
        full_master_connection_list = []
        for shard in self.shards:
            full_master_connection_list.append(shard.getConnection())
        return full_master_connection_list

    # Gets a cluster connection by key. On std redis the default connection is returned.
    def getConnectionByKey(self, key, command):
        clusterConn = self.getClusterConnection()
        target_node = clusterConn._determine_nodes(command, key) # we will always which will give us the node responsible for the key
        return clusterConn.get_redis_connection(target_node[0])

    def flush(self):
        self.getClusterConnection().flushall()

    def dumpAndReload(self, restart=False, shardId=None, timeout_sec=40):
        if shardId is None:
            for shard in self.shards:
                shard.dumpAndReload(restart=restart)
            self.waitCluster(timeout_sec=timeout_sec)
        else:
            self.shards[shardId -
                        1].dumpAndReload(restart=restart, shardId=None, timeout_sec=timeout_sec)

    def broadcast(self, *cmd):
        for shard in self.shards:
            shard.broadcast(*cmd)

    def checkExitCode(self):
        for shard in self.shards:
            if not shard.checkExitCode():
                return False
        return True

    def isUp(self):
        return self.envIsUp or self.envIsHealthy and self.waitCluster()

    def isHealthy(self):
        return self.envIsHealthy

    def isUnixSocket(self):
        return False

    def isTcp(self):
        return True

    def isTLS(self):
        return self.useTLS

    def exists(self, val):
        return self.getClusterConnection().exists(val)

    def hmset(self, *args):
        return self.getClusterConnection().hmset(*args)

    def keys(self, reg):
        return self.getClusterConnection().keys(reg)



import os
import sys
import time
import re
import copy
import redis
import itertools

def wait_for_conn(conn, proc, retries=20, command='PING', shouldBe=True):
    """Wait until a given Redis connection is ready"""
    err1 = ''
    while retries > 0:
        if proc.poll() is not None:
            raise Exception(f'Redis server is dead (pid={proc.pid})')
        try:
            if conn.execute_command(command) == shouldBe:
                return conn
        except redis.exceptions.BusyLoadingError:
            time.sleep(0.1)  # give extra 100msec in case of RDB loading
        except redis.ConnectionError as err:
            err1 = str(err)
        except redis.ResponseError as err:
            err1 = str(err)
            if not err1.startswith('DENIED'):
                raise
        time.sleep(0.1)
        retries -= 1
    raise Exception('Cannot establish connection %s: %s' % (conn, err1))

def expandBinary(binaryPath):
    if binaryPath is None:
        return binaryPath
    return os.path.expanduser(binaryPath) if binaryPath.startswith(
        '~/') else binaryPath

class Colors(object):
    @staticmethod
    def Cyan(data):
        return '\033[36m' + data + '\033[0m'

    @staticmethod
    def Yellow(data):
        return '\033[33m' + data + '\033[0m'

    @staticmethod
    def Bold(data):
        return '\033[1m' + data + '\033[0m'

    @staticmethod
    def Red(data):
        return '\033[31m' + data + '\033[0m'

    @staticmethod
    def Bred(data):
        return '\033[31;1m' + data + '\033[0m'

    @staticmethod
    def Gray(data):
        return '\033[30;1m' + data + '\033[0m'

    @staticmethod
    def Lgray(data):
        return '\033[30;47m' + data + '\033[0m'

    @staticmethod
    def Blue(data):
        return '\033[34m' + data + '\033[0m'

    @staticmethod
    def Green(data):
        return '\033[32m' + data + '\033[0m'

def fix_modules(modules, defaultModules=None):
    # modules is one of the following:
    # None
    # ['path',...]
    if modules:
        if not isinstance(modules, list):
            modules = [modules]
        modules = list(map(lambda p: os.path.abspath(p), modules))
    else:
        modules = defaultModules
    return modules

def split_by_semicolon(s):
    return list(filter(lambda s: s != '', map(lambda s: re.sub(r'\\(.)', r'\1', s.strip()), re.split(r'(?<!\\);', s))))

def args_list_to_dict(args_list):
    def dicty(args):
        return dict((seq.split(' ')[0], seq) for seq in args)
    return list(map(lambda args: dicty(args), args_list))

def join_lists(lists):
    return list(itertools.chain.from_iterable(lists))

def fix_modulesArgs(modules, modulesArgs, defaultArgs=None, haveSeqs=True):
    # modulesArgs is one of the following:
    # None
    # 'args ...': arg string for a single module
    # ['args ...', ...]: arg list for a single module
    # [['arg', ...', ...], ...]: arg strings for multiple modules

    # arg string is a string of words seperated by whitespace
    # arg string can be seperated by semicolons into (logical) arg lists.
    # semicolons can be escaped with a backslash.
    # arg list is a list of arg strings.
    # arg list starts with an arg name that can later be used for argument overriding.
    # arg strings are transformed into arg lists (haveSeqs parameter controls this behavior):
    # thus, 'num 1; names a b' becomes ['num 1', 'names a b']

    if type(modulesArgs) == str:
        # case # 'args ...': arg string for a single module
        # transformed into [['arg', ...]]
        modulesArgs = [split_by_semicolon(modulesArgs)]
    elif type(modulesArgs) == list:
        args = []
        is_list = False
        is_str = False
        for argx in modulesArgs:
            if type(argx) == list:
                # case [['arg', ...], ...]: arg strings for multiple modules
                # already transformed into [['arg', ...], ...]
                if is_str:
                    print(Colors.Bred('Error in args: %s' % str(modulesArgs)))
                    sys.exit(1)
                is_list = True
                if haveSeqs:
                    lists = map(lambda x: split_by_semicolon(x), argx)
                    args += [join_lists(lists)]
                else:
                    args += [argx]
            else:
                # case ['args ...', ...]: arg list for a single module
                # transformed into [['arg', ...], ...]
                if is_list:
                    print(Colors.Bred('Error in args: %s' % str(modulesArgs)))
                    sys.exit(1)
                is_str = True
                args += split_by_semicolon(argx)
        if is_str:
            args = [args]
        modulesArgs = args
    # modulesArgs is now [['arg', ...], ...]

    is_copy = not modulesArgs and defaultArgs
    if is_copy:
        modulesArgs = copy.deepcopy(defaultArgs)

    n = 0
    num_mods = len(modulesArgs) if modulesArgs else 0
    if defaultArgs:
        n = len(defaultArgs) - num_mods
        num_mods += n

    if isinstance(modules, list) and len(modules) > 1:
        n = len(modules) - num_mods

    if n > 0:
        if not modulesArgs:
            modulesArgs = []
        modulesArgs.extend([[]] * n)

    if is_copy or not defaultArgs:
        return modulesArgs

    # if there are fewer defaultArgs than modulesArgs, we should bail out
    # as we cannot pad the defaults with emply arg lists
    if defaultArgs and len(modulesArgs) > len(defaultArgs):
        print(Colors.Bred('Number of module args sets in Env does not match number of modules'))
        print(defaultArgs)
        print(modulesArgs)
        sys.exit(1)

    # for each module, sync defaultArgs to modulesARgs
    modules_args_dict = args_list_to_dict(modulesArgs)
    for imod, args_list in enumerate(defaultArgs):
        for arg in args_list:
            name = arg.split(' ')[0]
            if name not in modules_args_dict[imod]:
                modulesArgs[imod] += [arg]

    return modulesArgs



from redis import StrictRedis
from .exists_redis import ExistsRedisEnv


class Shard(StrictRedis):

    def __init__(self, cluster_address, port, password, cluster_credentials, **kwargs):
        self.port, self.node_id = port.split(":")
        self.password = password
        self.cluster_username, self.cluster_password = cluster_credentials.split(":")
        self.host = self._get_node_address(cluster_address)
        self.decodeResponses = kwargs.get('decodeResponses', False)
        StrictRedis.__init__(self, host = self.host, port = self.port, password = self.password,
                             decode_responses=self.decodeResponses, **kwargs)

    def dumpAndReload(self, restart=False, shardId=1, timeout_sec=0):
        self.save()
        try:
            self.execute_command('DEBUG', 'RELOAD')
        except redis.RedisError as err:
            raise err
    ###In the feature we want EnterpriseRedisClusterEnv will learn the cluster dynamicly
    ### So we will be able to run the suits without using operetto.
    def _get_node_address(self, cluster_address):
        base_url = 'https://{}:9443'.format(cluster_address)
        full_url ="{}/{}/{}".format(base_url, "v1/nodes", self.node_id)
        res = self.send_request('get', full_url)
        return res['addr']

    def send_request(self, method, url):
        import requests
        import urllib3
        urllib3.disable_warnings()
        responce = requests.api.request(method=method, url=url,
                             verify = False, auth=(self.cluster_username, self.cluster_password))
        if not responce.ok:
            raise Exception("Fail to send_request: responce code is {}".format(responce.status_code))
        return responce.json()

    def getMasterPort(self):
        return self.port

    def getConnection(self, shardId=1):
        return self

    def broadcast(self, *cmd):
        self.execute_command(*cmd)

class DB(StrictRedis):
    def __init__(self, host, port, password, shards_port, cluster_address, cluster_credentials, **kwargs):
        self.host = host
        self.port = port
        self.decodeResponses = kwargs.get('decodeResponses', False)
        self.shard_list = [Shard(cluster_address=cluster_address, 
                                 port=port, password=password, 
                                 cluster_credentials=cluster_credentials) for port in shards_port]
        StrictRedis.__init__(self, host = self.host, port = self.port,
                             decode_responses=self.decodeResponses, **kwargs)


class EnterpriseRedisClusterEnv(ExistsRedisEnv):

    def __init__(self, addr, cluster_address, cluster_credentials, password=None, shards_port=None, **kwargs):
        '''
        provide the ability to run tests on predefind RC
        usage:
        --existing-env-addr 'DB address {host}:{port}' --shards_ports "NODE_ID:INT,[NODE_ID:INT...]"
        --internal_password  "DB - internal_password" --cluster_address "master node IP"
        '''
        self.host, self.port = addr.split(':')
        self.cluster_address = cluster_address
        self.password = password
        self.database = DB(host=self.host, port=self.port,
                           password=password, shards_port = shards_port,
                           cluster_address = cluster_address,
                           cluster_credentials = cluster_credentials, **kwargs)
        self.shards = self.database.shard_list

    def waitCluster(self, timeout_sec=40):
        import time
        st = time.time()
        while st + timeout_sec > time.time():
            if all([shard.ping() for shard in self.shards]):
                return
            time.sleep(0.1)
        raise RuntimeError("Cluster OK wait loop timed out after %s seconds" % timeout_sec)

    def dumpAndReload(self, restart=False, shardId=None, timeout_sec=40):
        if shardId is None:
            for shard in self.shards:
                shard.dumpAndReload(restart=restart, timeout_sec=timeout_sec)
            self.waitCluster()
        else:
            self.shards[shardId - 1].dumpAndReload(restart=restart, shardId=None, timeout_sec=timeout_sec)

    def getClusterConnection(self):
        return self.getConnection()

    def getConnection(self, shardId=1):
        return self.database

    # List of nodes that initial bootstrapping can be done from
    def getMasterNodesList(self):
        full_master_list = []
        for shard in self.shards:
            node_info = {"host": None, "port": None, "unix_socket_path": None, "password": None}
            node_info["password"] = self.password
            node_info["host"] = 'localhost'
            node_info["port"] = shard.getMasterPort()
            full_master_list.append(node_info)
        return full_master_list

    # List containing a connection for each of the master nodes
    def getOSSMasterNodesConnectionList(self):
        full_master_connection_list = []
        for shard in self.shards:
            full_master_connection_list.append(shard.getConnection())
        return full_master_connection_list

    def isUp(self):
        return self.getConnection().ping()

    def isUnixSocket(self):
        return False

    def isTcp(self):
        return True




from __future__ import print_function
import redis
import subprocess
import sys
import os
import platform
from .utils import Colors, wait_for_conn


MASTER = 1
SLAVE = 2


class ExistsRedisEnv(object):
    def __init__(self, addr='localhost:6379', password = None, **kwargs):
        self.host, self.port = addr.split(':')
        self.port = int(self.port)
        self.password = password
        self.useTLS = kwargs['useTLS']
        self.decodeResponses = kwargs.get('decodeResponses', False)

    @property
    def has_interactive_debugger(self):
        return False

    def _printEnvData(self, prefix='', role=MASTER):
        print(Colors.Yellow(prefix + 'addr: %s:%d' % (self.host, self.port)))

    def printEnvData(self, prefix=''):
        print(Colors.Yellow(prefix + 'master:'))
        self._printEnvData(prefix + '\t', MASTER)

    def startEnv(self, masters = True, slaves = True):
        if not self.isUp():
            raise Exception('env is not up')

    def stopEnv(self, masters = True, slaves = True):
        pass

    def getConnection(self, shardId=1):
        return redis.StrictRedis(self.host, self.port, password=self.password, decode_responses=self.decodeResponses)

    def getSlaveConnection(self):
        raise Exception('asked for slave connection but no slave exists')

    # List of nodes that initial bootstrapping can be done from
    def getMasterNodesList(self):
        node_info = {"host": None, "port": None, "unix_socket_path": None, "password": None}
        node_info["password"] = self.password
        node_info["host"] = self.host
        node_info["port"] = self.port
        return [node_info]

    # List containing a connection for each of the master nodes
    def getOSSMasterNodesConnectionList(self):
        return [self.getConnection()]

    def getConnectionByKey(self, key, command):
        return self.getConnection()

    def _waitForBgsaveToFinish(self):
        # on new Redis version (6.2 and above)
        # flush trigger background rdb save
        # waiting for rdbsave to finish
        while True:
            if not self.getConnection().execute_command('info', 'Persistence')['rdb_bgsave_in_progress']:
                break

    def flush(self):
        self.getConnection().flushall()
        self._waitForBgsaveToFinish()

    def _waitForChild(self, conns):
        import time
        # Wait until file is available
        for con in conns:
            while True:
                info = con.info('persistence')
                if info['aof_rewrite_scheduled'] or info['aof_rewrite_in_progress']:
                    time.sleep(0.1)
                else:
                    break

    def dumpAndReload(self, restart=False, shardId=1, timeout_sec=0):
        self._waitForBgsaveToFinish()
        conn = self.getConnection()
        conn.save()
        try:
            conn.execute_command('DEBUG', 'RELOAD')
        except redis.RedisError as err:
            raise err

    def broadcast(self, *cmd):
        try:
            self.getConnection().execute_command(*cmd)
        except Exception as e:
            print(e)

    def checkExitCode(self):
        return True

    def isUp(self):
        return self.getConnection().ping()

    def isTLS(self):
        return self.useTLS

    def exists(self, val):
        return self.getConnection().exists(val)

    def hmset(self, *args):
        return self.getConnection().hmset(*args)

    def keys(self, reg):
        return self.getConnection().keys(reg)

    def isUnixSocket(self):
        return False

    def isTcp(self):
        return True



from __future__ import print_function
import subprocess
import redis
import os
from RLTest.utils import wait_for_conn, Colors


class CcsMock():
    CCS_UNIX_SOCKET_DEFAULT_PATH = '/tmp/ccs.sock'
    CCS_DB_RDB_FILE_NAME = 'syncer-pixie-ccs.rdb'
    CCS_LOG_FILE_NAME = 'ccs.log'

    def __init__(self, redisBinaryPath, proxyPort, directory=None, useSlaves=False, password=None, libPath=None):
        self.redisBinaryPath = os.path.expanduser(redisBinaryPath) if redisBinaryPath.startswith('~/') else redisBinaryPath
        self.useSlaves = useSlaves
        self.directory = directory
        self.generateArgs()
        self.libPath = os.path.expanduser(libPath) if libPath.startswith('~/') else libPath
        self.env = {}
        if self.libPath:
            self.env['LD_LIBRARY_PATH'] = self.libPath
        self.password = password
        self.proxyPort = proxyPort

    def generateArgs(self):
        self.args = self.redisBinaryPath.split()
        if self.directory:
            self.args += ['--dir', self.directory]
        self.args += ['--port', '0',
                      '--unixsocket', self.CCS_UNIX_SOCKET_DEFAULT_PATH,
                      '--dbfilename', self.CCS_DB_RDB_FILE_NAME,
                      '--logfile', self.CCS_LOG_FILE_NAME]

    def Start(self, shards, bdb_fields=None, endpoint_ccs_params=None, legacy_hash_slots=True, extra_keys=None):
        self.setup(shards, bdb_fields, endpoint_ccs_params, legacy_hash_slots, extra_keys)

    def Stop(self):
        if self.process:
            try:
                self.process.terminate()
                self.process.wait()
            except OSError:
                pass
            self.process = None

    def PrintEnvData(self, prefix=''):
        print(Colors.Yellow(prefix + 'pid: %d' % self.process.pid))
        print(Colors.Yellow(prefix + 'unix socket: %s' % self.CCS_UNIX_SOCKET_DEFAULT_PATH))
        print(Colors.Yellow(prefix + 'binary path: %s' % self.redisBinaryPath))
        if self.directory:
            print(Colors.Yellow(prefix + 'db dir path: %s' % (self.directory)))
        print(Colors.Yellow(prefix + 'rdb file name: %s' % (self.CCS_DB_RDB_FILE_NAME)))
        print(Colors.Yellow(prefix + 'log file name: %s' % (self.CCS_LOG_FILE_NAME)))
        if self.libPath:
            print(Colors.Yellow(prefix + 'lib path: %s' % (self.libPath)))

    def setup(self, shards, bdb_fields=None, endpoint_ccs_params=None, legacy_hash_slots=True, extra_keys=None):

        self.process = subprocess.Popen(stdout=subprocess.PIPE, stderr=subprocess.STDOUT, args=self.args, env=self.env)

        # Create fake cluster, node and dmc objects
        self.conn = redis.Redis(unix_socket_path=self.CCS_UNIX_SOCKET_DEFAULT_PATH)
        wait_for_conn(self.conn)

        self.conn.hmset('cluster', {'name': 'fakecluster'})
        self.conn.hmset('node:1', {'addr': '127.0.0.1',
                                   'dmc_uid': '1'})
        self.conn.sadd('node:all', '1')
        self.conn.hmset('dmc:1', {'threads': '1',
                                  'max_threads': '1',
                                  'log_level': 'info'})
        self.conn.sadd('dmc:all', '1')
        # self.conn.hmset('cluster_cert', {"syncer_cert": "syncer_cert", "syncer_key": "syncer_key"}) todo: check if needed
        self.disperse_shards = False
        self.next_node_uid = 1

        self.create_bdb(shards, bdb_fields, endpoint_ccs_params, legacy_hash_slots, extra_keys)

    def set_server_ccs_values(self, serverId, port, assigned_slots, bdb_uid, role):
        """
        Set all the neccesary values of a server in the ccs
        """
        redis_uid = serverId
        self.conn.hmset('redis:%s' % str(redis_uid),
                        {'port': port,
                         'node_uid': str(self.next_node_uid),
                         'bdb_uid': bdb_uid,
                         'role': role,
                         'assigned_slots': assigned_slots})
        self.conn.sadd('redis:all', redis_uid)
        # we promote the node uid just in case of master shard because we want disperion of master shards
        if role == 'master' and self.disperse_shards:
            n_nodes = len(self.conn.smembers('node:all'))
            self.next_node_uid = self.next_node_uid % n_nodes + 1

    def update_bdb_redis_list(self, bdb_uid, shards):
        """
        Update the redis list in the ccs of the given bdb
        """
        bdb_uid = str(bdb_uid)
        redis_uids = []
        for shard in shards:
            redis_uids.append(str(shard.masterServerId))
            if self.useSlaves:
                redis_uids.append(str(shard.slaveServerId))
        self.conn.hset('bdb:%s' % bdb_uid, 'redis_list', ','.join(redis_uids))

    def create_bdb(self, shards, bdb_fields=None, endpoint_ccs_params=None, legacy_hash_slots=True, extra_keys=None):
        bdb_uid = 1
        port = self.proxyPort

        if legacy_hash_slots:
            number_of_slots_in_bdb = 4096
            translation_offset = 1
        else:
            number_of_slots_in_bdb = 16384
            translation_offset = 0
        slots_per_shard = float(number_of_slots_in_bdb) / len(shards)
        first_slot = translation_offset
        last_slot = number_of_slots_in_bdb - 1 + translation_offset

        start_slots = [int(round(first_slot + i * slots_per_shard)) for i in range(len(shards))]
        end_slots = [next_start_slot - 1 for next_start_slot in start_slots[1:]] + [last_slot]
        slots = ["%s-%s" % (pair[0], pair[1]) for pair in zip(start_slots, end_slots)]

        for index, shard in enumerate(shards):
            self.set_server_ccs_values(shard.masterServerId, shard.port, str(slots[index]), bdb_uid, 'master')

            if self.useSlaves:
                self.set_server_ccs_values(shard.slaveServerId, shard.GetSlavePort(), str(slots[index]), bdb_uid, 'slave')

        self.conn.hmset('bdb:%s' % bdb_uid,
                        {'shards_count': len(shards),
                         'type': 'redis',
                         'endpoint_list': '%s:1' % bdb_uid,
                         'replication': ('enabled' if self.useSlaves else 'disabled'),
                         'authentication_admin_pass': 'secret',
                         'internal_pass': self.password,
                         'redis_version': '4.0',
                         'implicit_shard_key': 'enabled'})

        if legacy_hash_slots:
            self.conn.hmset('bdb:%s' % bdb_uid,
                            {'hash_slots_policy': 'legacy',
                             'shard_function': 'crc12'})
        else:
            self.conn.hmset('bdb:%s' % bdb_uid,
                            {'hash_slots_policy': '16k'})

        if bdb_fields:
            self.conn.hmset('bdb:%s' % bdb_uid, bdb_fields)
            # if 'ssl' in bdb_fields and bdb_fields['ssl'] in ["enabled", "replica_ssl"]:
            #     with open(CLIENT_CERT_FILE, 'r') as cert_file:
            #         cert = cert_file.readlines()
            #     self.conn.hmset('bdb:%s' % bdb_uid,
            #                     {'authentication_ssl_client_certs': "".join(cert)})

        if extra_keys:
            for key, val in extra_keys.items():
                self.conn.hmset(key, val)

        # Set up endpoint
        proxy_policy = self.conn.hget('bdb:%s' % bdb_uid, 'proxy_policy')
        if not proxy_policy:
            proxy_policy = 'all-nodes'

        self.conn.hmset('endpoint:%s:1' % bdb_uid,
                        {'bdb_uid': bdb_uid,
                         'port': port,
                         'proxy_policy': proxy_policy})
        self.conn.hmset('endpoint:%s:1' % bdb_uid,
                        {'bdb_uid': bdb_uid,
                         'port': port,
                         'proxy_policy': proxy_policy,
                         'proxy_uids': '1'})

        if endpoint_ccs_params:
            self.conn.hmset('endpoint:%s:1' % bdb_uid, endpoint_ccs_params)

        self.conn.sadd('endpoint:all', '%s:1' % bdb_uid)

        self.conn.sadd('bdb:all', bdb_uid)
        self.update_bdb_redis_list(bdb_uid, shards)



from __future__ import print_function

from RLTest.redis_std import StandardEnv
from RLTest.utils import Colors, wait_for_conn
from .CcsMock import CcsMock
from .Dmc import Dmc
import redis
import os
import json
from zipfile import ZipFile


SHARD_PASSWORD = 'password'


class EnterpriseClusterEnv():
    DMC_PORT = 10000
    MODULE_WORKING_DIR = '/tmp/'

    def __init__(self, **kwargs):

        self.shards = []
        self.envIsUp = False
        self.envIsHealthy = False
        self.modulePath = kwargs.pop('modulePath')
        self.moduleArgs = kwargs['moduleArgs']
        self.shardsCount = kwargs.pop('shardsCount')
        self.dmcBinaryPath = kwargs.pop('dmcBinaryPath')
        useSlaves = kwargs.get('useSlaves', False)

        self.preperModule()
        startPort = 20000
        totalRedises = self.shardsCount * (2 if useSlaves else 1)
        for i in range(0, totalRedises, (2 if useSlaves else 1)):
            shard = StandardEnv(port=startPort, serverId=(i + 1), password=SHARD_PASSWORD, modulePath=self.moduleSoFilePath, **kwargs)
            self.shards.append(shard)
            startPort += 2

        self.ccs = CcsMock(redisBinaryPath=kwargs['redisBinaryPath'], directory=kwargs['dbDirPath'], useSlaves=kwargs['useSlaves'],
                           password=SHARD_PASSWORD, proxyPort=self.DMC_PORT, libPath=kwargs['libPath'])
        self.dmc = Dmc(directory=kwargs['dbDirPath'], dmcBinaryPath=self.dmcBinaryPath, libPath=kwargs['libPath'])
        self.envIsUp = False

    def preperModule(self):
        if self.modulePath is None:
            self.moduleSoFilePath = None
            self.moduleConfig = None
            return
        if not self.modulePath.endswith('zip'):
            raise Exception('module on enterprise cluster must be a zip file')
        with ZipFile(self.modulePath, 'r') as myzip:
            moduleJson = [a.filename for a in myzip.infolist() if a.filename.endswith('.json')]
            if len(moduleJson) != 1:
                self.moduleSoFilePath = None
                self.moduleConfig = None
                return
            moduleJson = moduleJson[0]
            myzip.extractall(self.MODULE_WORKING_DIR)
            with open(os.path.join(self.MODULE_WORKING_DIR, moduleJson), 'rt') as f:
                self.moduleConfig = json.load(f)
            self.moduleSoFilePath = os.path.join(self.MODULE_WORKING_DIR, self.moduleConfig['module_file'])
            if self.moduleArgs is None:
                self.moduleArgs = self.moduleConfig['command_line_args']

    def printEnvData(self, prefix=''):
        print(Colors.Yellow(prefix + 'bdb info:'))
        print(Colors.Yellow(prefix + '\tlistening port:%d' % self.DMC_PORT))
        print(Colors.Yellow(prefix + '\tshards count:%d' % len(self.shards)))
        if self.modulePath:
            print(Colors.Yellow(prefix + '\tzip module path:%s' % self.modulePath))
        if self.moduleSoFilePath:
            print(Colors.Yellow(prefix + '\tso module path:%s' % self.moduleSoFilePath))
        if self.moduleArgs:
            print(Colors.Yellow(prefix + '\tmodule args:%s' % self.moduleArgs))
        for i, shard in enumerate(self.shards):
            print(Colors.Yellow(prefix + 'shard: %d' % (i + 1)))
            shard.printEnvData(prefix + '\t')
        print(Colors.Yellow(prefix + 'ccs:'))
        self.ccs.PrintEnvData(prefix + '\t')
        print(Colors.Yellow(prefix + 'dmc:'))
        self.dmc.PrintEnvData(prefix + '\t')

    def startEnv(self, masters = True, slaves = True):
        if self.envIsUp:
            return  # env is already up
        for shard in self.shards:
            shard.startEnv(masters, slaves)

        ccs_bdb_config = {'shard_key_regex': '012.*\{(?<tag>.*)\}.*00a(?<tag>.*)',
                          'sharding': 'enabled' if self.shardsCount > 0 else 'disabled'}

        extra_keys = {}
        if self.moduleConfig:
            ccs_bdb_config['module_list'] = self.moduleConfig['module_name']
            extra_keys['module:%s' % self.moduleConfig['module_name']] = {'module_name': self.moduleConfig['module_name'],
                                                                          'commands': ','.join([c['command_name'] for c in self.moduleConfig['commands']])}
            for c in self.moduleConfig['commands']:
                key_name = 'module_command:%s' % c['command_name']
                extra_keys[key_name] = {}
                for key, val in c.items():
                    extra_keys[key_name][key] = val

        self.ccs.Start(self.shards, bdb_fields=ccs_bdb_config, legacy_hash_slots=False, extra_keys=extra_keys)
        self.dmc.Start()
        con = self.getConnection()
        wait_for_conn(con, command='sping', shouldBe=['SPONG 0' for i in self.shards])
        self.envIsUp = True
        self.envIsHealthy = True

    def stopEnv(self, masters = True, slaves = True):
        for shard in self.shards:
            shard.stopEnv(masters, slaves)
            self.envIsUp = self.envIsUp or shard.envIsUp
            self.envIsHealthy = self.envIsHealthy and shard.envIsUp
        self.ccs.Stop()
        self.dmc.Stop()

    def getConnection(self, shardId=1):
        return redis.Redis('localhost', self.DMC_PORT)

    def getSlaveConnection(self):
        raise Exception('unsupported')

    def flush(self):
        self.getConnection().flushall()

    def dumpAndReload(self, restart=False, shardId=None, timeout_sec=40):
        if shardId is None:
            for shard in self.shards:
                shard.dumpAndReload(restart=restart)
            self.dmc.Stop()
            self.dmc.Start()
            con = self.getConnection()
            wait_for_conn(con, command='sping', shouldBe=['SPONG 0' for i in self.shards])
        else:
            self.shards[shardId - 1].dumpAndReload(restart=restart, shardId=None)

    def broadcast(self, *cmd):
        for shard in self.shards:
            shard.broadcast(*cmd)

    def checkExitCode(self):
        for shard in self.shards:
            if not shard.checkExitCode():
                return False
        return True

    def exists(self, val):
        return self.getConnection().exists(val)

    def hmset(self, *args):
        return self.getConnection().hmset(*args)

    def keys(self, reg):
        return self.getConnection().keys(reg)

    def isUp(self):
        raise Exception('unsupported operation')



from __future__ import print_function
import os.path
import shutil
import distro
import subprocess
import sys

from RLTest.utils import Colors


OS_NAME = distro.linux_distribution()[2]
REPO_ROOT = os.path.expanduser('~/.RLTest')
ENTERPRISE_VERSION = '5.2.0'
ENTERPRISE_SUB_VERSION = '14'
ENTERPRISE_TAR_FILE_NAME = 'redislabs-%s-%s-%s-amd64.tar' % (
    ENTERPRISE_VERSION, ENTERPRISE_SUB_VERSION, OS_NAME)
ENTERPRISE_URL = 'https://s3.amazonaws.com/rlec-downloads/%s/%s' % (
    ENTERPRISE_VERSION, ENTERPRISE_TAR_FILE_NAME)
DEBIAN_PKG_NAME = 'redislabs_%s-%s~%s_amd64.deb' % (
        ENTERPRISE_VERSION, ENTERPRISE_SUB_VERSION, OS_NAME)


class BinaryRepository(object):
    """
    Installation manager for Redis Enterprise.
    The long-term idea behind this is to allow for managing different versions, etc.
    but for now it's just here to cut some code out of the main executable
    """
    def __init__(self, root=REPO_ROOT, url=ENTERPRISE_URL, debname=DEBIAN_PKG_NAME):
        self.root = root
        self.url = url
        self.debname = debname

    def download_binaries(self, binariesName='binaries.tar'):
        print(Colors.Yellow('installing enterprise binaries'))
        print(Colors.Yellow('creating RLTest working dir: %s' % self.root))
        try:
            shutil.rmtree(self.root)
            os.makedirs(self.root)
        except Exception:
            pass

        print(Colors.Yellow('download binaries'))
        args = ['wget', self.url, '-O', os.path.join(self.root, binariesName)]
        process = subprocess.Popen(args=args, stdout=sys.stdout,
                                        stderr=sys.stdout)
        process.wait()
        if process.poll() != 0:
            raise Exception('failed to download enterprise binaries from s3')

        print(Colors.Yellow('extracting binaries'))

        args = ['tar', '-xvf', os.path.join(self.root, binariesName),
                '--directory', self.root, self.debname]
        process = subprocess.Popen(args=args, stdout=sys.stdout, stderr=sys.stdout)
        process.wait()
        if process.poll() != 0:
            raise Exception(
                'failed to extract binaries to %s' % self.self.root)

        # TODO: Support centos that does not have dpkg command
        args = ['dpkg', '-x', os.path.join(self.root, self.debname),
                self.root]
        process = subprocess.Popen(args=args, stdout=sys.stdout,
                                        stderr=sys.stdout)
        process.wait()
        if process.poll() != 0:
            raise Exception(
                'failed to extract binaries to %s' % self.self.root)

        print(Colors.Yellow('finished installing enterprise binaries'))


# coding=utf-8
from __future__ import print_function

import contextlib
import inspect
import os
import sys
import unittest
import warnings

from .exists_redis import ExistsRedisEnv
from .redis_cluster import ClusterEnv
from .redis_enterprise_cluster import EnterpriseRedisClusterEnv
from .redis_std import StandardEnv
from .utils import Colors, expandBinary, fix_modules, fix_modulesArgs
from packaging import version


class TestAssertionFailure(Exception):
    pass


def genDeprecated(name, target):
    def method(*argc, **nargs):
        warnings.warn('%s is deprecated, use %s instead' % (str(name), str(target)), DeprecationWarning)
        return target(*argc, **nargs)
    return method


class Query:
    def __init__(self, env, *query, **options):
        self.query = query
        self.options = options
        self.env = env
        self.errorRaised = False
        self._evaluate()

    def _evaluate(self):
        try:
            self.res = self.env.cmd(*self.query, **self.options)
        except Exception as e:
            self.res = str(e)
            self.errorRaised = True

    def _prettyPrint(self, result, prefix='\t'):
        if type(result) is list:
            print(prefix + '[')
            for r in result:
                self._prettyPrint(r, prefix + '\t')
            print(prefix + ']')
            return
        print(prefix + str(result))

    def prettyPrint(self):
        self._prettyPrint(self.res)
        return self

    def debugPrint(self):
        self.env.debugPrint('query: %s, result: %s' % (self.query, self.res), force=True)
        return self

    def apply(self, fn):
        self.res = fn(self.res)
        return self

    def map(self, fn):
        self.res = list(map(fn, self.res))
        return self

    def equal(self, expected):
        self.env.assertEqual(self.res, expected, 1)
        return self

    def noEqual(self, expected):
        self.env.assertNotEqual(self.res, expected, 1)
        return self

    def true(self):
        self.env.assertTrue(self.res, 1)
        return self

    def false(self):
        self.env.assertFalse(self.res, 1)
        return self

    def ok(self):
        self.env.assertEqual(self.res, 'OK', 1)
        return self

    def contains(self, val):
        self.env.assertContains(val, self.res, 1)
        return self

    def notContains(self, val):
        self.env.assertNotContains(val, self.res, 1)
        return self

    def error(self):
        self.env.assertTrue(self.errorRaised, 1)
        return self

    def noError(self):
        self.env.assertFalse(self.errorRaised, 1)
        return self

    raiseError = genDeprecated('raiseError', error)
    notRaiseError = genDeprecated('notRaiseError', noError)


class Defaults:
    module = None
    module_args = None

    env = 'oss'
    env_factory = lambda *args, **kwargs: Env(*args, **kwargs)
    binary = 'redis-server'
    proxy_binary = None
    re_binary = None
    re_libdir = None
    decode_responses = False
    use_aof = False
    use_rdb_preamble = True
    use_TLS = False
    tls_cert_file = None
    tls_key_file = None
    tls_ca_cert_file = None
    tls_passphrase = None
    debugger = None
    sanitizer = None
    debug_print = False
    debug_pause = False
    no_capture_output = False
    print_verbose_information_on_failure = False
    no_log = False
    exit_on_failure = False
    verbose = 0
    logdir = None
    loglevel = None
    use_slaves = False
    num_shards = 1
    external_addr = 'localhost:6379'
    use_unix = False
    randomize_ports = False
    oss_password = None
    cluster_node_timeout = None
    curr_test_name = None
    port=6379
    enable_debug_command=False
    enable_protected_configs=False
    terminate_retries=None
    terminate_retry_secs=None
    protocol=2

    def getKwargs(self):
        kwargs = {
            'modulePath': self.module,
            'moduleArgs': self.module_args,
            'port': self.port,
            'useSlaves': self.use_slaves,
            'useAof': self.use_aof,
            'useRdbPreamble': self.use_rdb_preamble,
            'dbDirPath': self.logdir,
            'debugger': self.debugger,
            'sanitizer': self.sanitizer,
            'noCatch': self.no_capture_output,
            'noLog': self.no_log,
            'verbose': self.verbose,
            'useTLS': self.use_TLS,
            'tlsCertFile': self.tls_cert_file,
            'tlsKeyFile': self.tls_key_file,
            'tlsCaCertFile': self.tls_ca_cert_file,
            'tlsPassphrase': self.tls_passphrase,
            'password': self.oss_password,
            'terminateRetries': self.terminate_retries,
            'terminateRetrySecs': self.terminate_retry_secs,
        }
        return kwargs


class Env:
    RTestInstance = None
    EnvCompareParams = ['module', 'moduleArgs', 'env', 'useSlaves', 'shardsCount', 'useAof',
                        'useRdbPreamble', 'forceTcp', 'enableDebugCommand', 'enableProtectedConfigs', 'protocol']

    def compareEnvs(self, env):
        if env is None:
            return False
        for param in Env.EnvCompareParams:
            if self.__dict__[param] != env.__dict__[param]:
                return False
        return True

    def __init__(self, testName=None, testDescription=None, module=None,
                 moduleArgs=None, env=None, useSlaves=None, shardsCount=None, decodeResponses=None,
                 useAof=None, useRdbPreamble=None, forceTcp=False, useTLS=False, tlsCertFile=None, tlsKeyFile=None,
                 tlsCaCertFile=None, tlsPassphrase=None, logDir=None, redisBinaryPath=None, dmcBinaryPath=None,
                 redisEnterpriseBinaryPath=None, noDefaultModuleArgs=False, clusterNodeTimeout = None,
                 freshEnv=False, enableDebugCommand=None, enableProtectedConfigs=None, protocol=None,
                 terminateRetries=None, terminateRetrySecs=None):

        self.testName = testName if testName else Defaults.curr_test_name
        if self.testName is None:
            self.testName = '%s.%s' % (inspect.getmodule(inspect.currentframe().f_back).__name__, inspect.currentframe().f_back.f_code.co_name)
        self.testName = self.testName.replace(' ', '_')

        if testDescription:
            print(Colors.Gray('\tdescription: ' + testDescription))

        self.module = fix_modules(module, Defaults.module)
        if noDefaultModuleArgs:
            self.moduleArgs = fix_modulesArgs(self.module, moduleArgs)
        else:
            self.moduleArgs = fix_modulesArgs(self.module, moduleArgs, Defaults.module_args)
        self.env = env if env else Defaults.env
        self.useSlaves = useSlaves if useSlaves else Defaults.use_slaves
        self.shardsCount = shardsCount if shardsCount else Defaults.num_shards
        self.decodeResponses = decodeResponses if decodeResponses else Defaults.decode_responses
        self.useAof = useAof if useAof else Defaults.use_aof
        self.useRdbPreamble = useRdbPreamble if useRdbPreamble is not None else Defaults.use_rdb_preamble
        self.verbose = Defaults.verbose
        self.logDir = logDir if logDir else Defaults.logdir
        self.forceTcp = forceTcp
        self.debugger = Defaults.debugger
        self.sanitizer = Defaults.sanitizer
        self.useTLS = useTLS if useTLS else Defaults.use_TLS
        self.tlsCertFile = tlsCertFile if tlsCertFile else Defaults.tls_cert_file
        self.tlsKeyFile = tlsKeyFile if tlsKeyFile else Defaults.tls_key_file
        self.tlsCaCertFile = tlsCaCertFile if tlsCaCertFile else Defaults.tls_ca_cert_file
        self.tlsPassphrase = tlsPassphrase if tlsPassphrase else Defaults.tls_passphrase

        self.redisBinaryPath = expandBinary(redisBinaryPath) if redisBinaryPath else Defaults.binary
        self.dmcBinaryPath = expandBinary(dmcBinaryPath) if dmcBinaryPath else Defaults.proxy_binary
        self.redisEnterpriseBinaryPath = expandBinary(redisEnterpriseBinaryPath) if redisEnterpriseBinaryPath else Defaults.re_binary
        self.clusterNodeTimeout = clusterNodeTimeout if clusterNodeTimeout else Defaults.cluster_node_timeout
        self.port = Defaults.port
        self.enableDebugCommand = enableDebugCommand if enableDebugCommand else Defaults.enable_debug_command
        self.enableProtectedConfigs = enableProtectedConfigs if enableProtectedConfigs\
            else Defaults.enable_protected_configs
        self.terminateRetries = terminateRetries
        self.terminateRetrySecs = terminateRetrySecs

        self.protocol = protocol if protocol is not None else Defaults.protocol

        self.assertionFailedSummary = []

        if not freshEnv and Env.RTestInstance and Env.RTestInstance.currEnv and self.compareEnvs(Env.RTestInstance.currEnv):
            self.envRunner = Env.RTestInstance.currEnv.envRunner
        else:
            if Env.RTestInstance and Env.RTestInstance.currEnv:
                Env.RTestInstance.currEnv.stop()
            self.envRunner = self.getEnvByName()

        try:
            os.makedirs(self.logDir)
        except Exception:
            pass

        self.start()
        if self.verbose >= 2:
            print(Colors.Blue('\tenv data:'))
            self.envRunner.printEnvData('\t\t')

        if Env.RTestInstance:
            Env.RTestInstance.currEnv = self

        if Defaults.debug_pause:
            input('\tenv is up, attach to any process with gdb and press any button to continue.')

    def getInformationBeforeDispose(self):
        return {
            "env": self.env,
            "test": self.testName,
            "env_info": self.envRunner.getInformationBeforeDispose()
        }

    def getInformationAfterDispose(self):
        return self.envRunner.getInformationAfterDispose()

    def getEnvByName(self):
        verbose = False
        kwargs = self.getEnvKwargs()
        single_args = self.getSingleArgs()

        test_fname = self.testName.replace(':', '_')

        if self.env == 'oss':
            kwargs.update(single_args)
            kwargs['password'] = Defaults.oss_password
            return StandardEnv(redisBinaryPath=self.redisBinaryPath,
                               outputFilesFormat='%s-' + '%s-oss' % test_fname,
                               **kwargs)
        if self.env == 'enterprise':
            kwargs.update(single_args)
            kwargs['libPath'] = Defaults.re_libdir
            return StandardEnv(redisBinaryPath=self.redisEnterpriseBinaryPath,
                               outputFilesFormat='%s-' + '%s-oss' % test_fname,
                               **kwargs)
        if self.env == 'enterprise-cluster':
            kwargs['libPath'] = Defaults.re_libdir
            return EnterpriseClusterEnv(shardsCount=self.shardsCount,
                                        redisBinaryPath=self.redisEnterpriseBinaryPath,
                                        outputFilesFormat='%s-' + '%s-re-cluster' % test_fname,
                                        dmcBinaryPath=Defaults.proxy_binary,
                                        **kwargs)
        if self.env == 'oss-cluster':
            kwargs['password'] = Defaults.oss_password
            return ClusterEnv(shardsCount=self.shardsCount, redisBinaryPath=self.redisBinaryPath,
                              outputFilesFormat='%s-' + '%s-oss-cluster' % test_fname,
                              randomizePorts=Defaults.randomize_ports,
                              **kwargs)

        if self.env == 'existing-env':
            return ExistsRedisEnv(addr=Defaults.external_addr, **kwargs)

        if self.env == 'cluster_existing-env':
            return EnterpriseRedisClusterEnv(addr = Defaults.external_addr, password = Defaults.internal_password,
                                             shards_port=Defaults.shards_ports,
                                             cluster_address = Defaults.cluster_address,
                                             cluster_credentials= Defaults.cluster_credentials, **kwargs)

    def getSingleArgs(self):
        single_args = {}
        if Defaults.randomize_ports:
            single_args['port'] = 0
        if Defaults.use_unix:
            single_args['unix'] = True
        if self.forceTcp and self.env != 'existing-env':
            single_args['port'] = 0
            single_args.pop('unix', None)
        return single_args

    def getEnvKwargs(self):
        kwargs = {
            'modulePath': self.module,
            'moduleArgs': self.moduleArgs,
            'useSlaves': self.useSlaves,
            'decodeResponses': self.decodeResponses,
            'useAof': self.useAof,
            'useRdbPreamble': self.useRdbPreamble,
            'dbDirPath': self.logDir,
            'loglevel': Defaults.loglevel,
            'debugger': Defaults.debugger,
            'sanitizer': Defaults.sanitizer,
            'noCatch': Defaults.no_capture_output,
            'noLog': Defaults.no_log,
            'verbose': Defaults.verbose,
            'useTLS': self.useTLS,
            'tlsCertFile': self.tlsCertFile,
            'tlsKeyFile': self.tlsKeyFile,
            'tlsCaCertFile': self.tlsCaCertFile,
            'clusterNodeTimeout': self.clusterNodeTimeout,
            'tlsPassphrase': self.tlsPassphrase,
            'port': self.port,
            'enableDebugCommand': self.enableDebugCommand,
            'enableProtectedConfigs': self.enableProtectedConfigs,
            'protocol': self.protocol,
            'terminateRetries': self.terminateRetries,
            'terminateRetrySecs': self.terminateRetrySecs,
        }
        return kwargs

    def start(self, masters = True, slaves = True ):
        self.envRunner.startEnv(masters, slaves)
        self.con = self.getConnection()

    def stop(self, masters = True, slaves = True):
        self.envRunner.stopEnv(masters, slaves)

    def getEnvStr(self):
        return self.env

    def getConnection(self, shardId=1):
        return self.envRunner.getConnection(shardId)

    def getClusterConnectionIfNeeded(self):
        if isinstance(self.envRunner, ClusterEnv):
            return self.envRunner.getClusterConnection()
        elif isinstance(self.envRunner, EnterpriseRedisClusterEnv):
            return self.envRunner.getClusterConnection()
        else:
            return self.getConnection()

    def getSlaveConnection(self):
        return self.envRunner.getSlaveConnection()

    # List of nodes that initial bootstrapping can be done from
    def getMasterNodesList(self):
        return self.envRunner.getMasterNodesList()

    # List containing a connection for each of the master nodes
    def getOSSMasterNodesConnectionList(self):
        return self.envRunner.getOSSMasterNodesConnectionList()

    def getConnectionByKey(self, key, command):
        return self.envRunner.getConnectionByKey(key, command)

    def flush(self):
        self.envRunner.flush()

    def isCluster(self):
        return 'cluster' in self.env or os.getenv("RLEC_CLUSTER") == "1"

    def isEnterpiseCluster(self):
        return isinstance(self.envRunner, EnterpriseRedisClusterEnv)

    def isDebugger(self):
        return self.debugger is not None

    def _getCallerPosition(self, back_frames):
        frame = inspect.currentframe()
        while frame and back_frames > 0:
            back_frames -= 1
            frame = frame.f_back
        if frame:
            return '%s:%s' % (
                os.path.basename(frame.f_code.co_filename),
                frame.f_lineno)

    def _assertion(self, checkStr, trueValue, depth=0, message=None):
        basemsg = Colors.Yellow(checkStr) + '\t' + Colors.Gray(self._getCallerPosition(3 + depth))
        if message:
            basemsg += ' [{}]'.format(message)

        if trueValue and self.verbose:
            print('\t' + Colors.Green('✅  (OK):\t') + basemsg)
        elif not trueValue:
            failureSummary = Colors.Bred('❌  (FAIL):\t') + basemsg
            print('\t' + failureSummary)
            if Defaults.exit_on_failure:
                raise TestAssertionFailure('Assertion Failed!')

            self.assertionFailedSummary.append(failureSummary)

    def getNumberOfFailedAssertion(self):
        return len(self.assertionFailedSummary)

    def assertEqual(self, first, second, depth=0, message=None):
        self._assertion('%s == %s' % (repr(first), repr(second)), first == second, depth, message=message)

    def assertNotEqual(self, first, second, depth=0, message=None):
        self._assertion('%s != %s' % (repr(first), repr(second)), first != second, depth, message=message)

    def assertOk(self, val, depth=0, message=None):
        self.assertEqual(val, 'OK', depth + 1, message=message)

    def assertTrue(self, val, depth=0, message=None):
        self.assertEqual(bool(val), True, depth + 1, message=message)

    def assertFalse(self, val, depth=0, message=None):
        self.assertEqual(bool(val), False, depth + 1, message=message)

    def assertContains(self, value, holder, depth=0):
        self._assertion('%s should contain %s' % (repr(holder), repr(value)), value in holder, depth)

    def assertNotContains(self, value, holder, depth=0):
        self._assertion('%s should not contain %s' % (repr(holder), repr(value)), value not in holder, depth)

    def assertGreaterEqual(self, value1, value2, depth=0):
        self._assertion('%s >= %s' % (repr(value1), repr(value2)), value1 >= value2, depth)

    def assertGreater(self, value1, value2, depth=0):
        self._assertion('%s > %s' % (repr(value1), repr(value2)), value1 > value2, depth)

    def assertLessEqual(self, value1, value2, depth=0):
        self._assertion('%s <= %s' % (repr(value1), repr(value2)), value1 <= value2, depth)

    def assertLess(self, value1, value2, depth=0):
        self._assertion('%s < %s' % (repr(value1), repr(value2)), value1 < value2, depth)

    def assertIsNotNone(self, value, depth=0):
        self._assertion('%s is not None' % (repr(value)), value is not None, depth)

    def assertIsNone(self, value, depth=0):
        self._assertion('%s is None' % (repr(value)), value is None, depth)

    def assertIsInstance(self, value, instance, depth=0):
        self._assertion('%s instance of %s' % (repr(value), repr(instance)), isinstance(value, instance), depth)

    def assertAlmostEqual(self, value1, value2, delta, depth=0, message=None):
        self._assertion('%s almost equels %s (delta %s)' % (repr(value1), repr(value2), repr(delta)), abs(value1 - value2) <= delta, depth, message)

    def expect(self, *query, **options):
        return Query(self, *query, **options)

    def cmd(self, *query, **options):
        res = self.con.execute_command(*query, **options)
        self.debugPrint('query: %s, result: %s' % (repr(query), repr(res)))
        return res

    def assertCmdOk(self, cmd, *args, **kwargs):
        self.assertOk(self.cmd(cmd, *args, **kwargs))

    def exists(self, val):
        warnings.warn("Exists is deprecated, use cmd instead", DeprecationWarning)
        return self.envRunner.exists(val)

    def assertExists(self, val, depth=0):
        warnings.warn("AssertExists is deprecated, use cmd instead", DeprecationWarning)
        self._assertion('%s exists in db' % repr(val), self.con.exists(val), depth=0)

    def executeCommand(self, *query, **options):
        warnings.warn("execute_command is deprecated, use cmd instead", DeprecationWarning)
        return self.cmd(*query, **options)

    def reloadingIterator(self):
        yield 1
        self.dumpAndReload()
        yield 2

    def dumpAndReload(self, restart=False, shardId=None, timeout_sec=40):
        self.envRunner.dumpAndReload(restart=restart, shardId=shardId, timeout_sec=timeout_sec)

    def hmset(self, *args):
        warnings.warn("hmset is deprecated, use Cmd instead", DeprecationWarning)
        return self.envRunner.hmset(*args)

    def keys(self, reg):
        warnings.warn("keys is deprecated, use Cmd instead", DeprecationWarning)
        return self.envRunner.keys(reg)

    def assertRaises(self, var1, var2, *query):
        warnings.warn("assertRaises is deprecated, use Expect + RaiseError instead", DeprecationWarning)
        self.expect(*query).raiseError()

    @contextlib.contextmanager
    def assertResponseError(self, msg=None, contained=None):
        """
        Assert that a context block with a redis command triggers a redis error response.

        For Example:

            with self.assertResponseError():
                r.execute_command('non_existing_command')
        """

        warnings.warn("assertResponseError is deprecated, use Expect + RaiseError instead", DeprecationWarning)

        try:
            yield 1
        except Exception as e:
            if contained:
                self.assertContains(contained, str(e), depth=2)
            self._assertion('Expected Response Error', True, depth=1)
        else:
            self._assertion('Expected Response Error', False, depth=1)

    def restartAndReload(self, shardId=None, timeout_sec=40):
        self.dumpAndReload(restart=True, shardId=shardId, timeout_sec=timeout_sec)

    def broadcast(self, *cmd):
        self.envRunner.broadcast(*cmd)

    def debugPrint(self, msg, force=False):
        if Defaults.debug_print or force:
            print('\t' + Colors.Bold('debug:\t') + Colors.Gray(msg))

    def checkExitCode(self):
        return self.envRunner.checkExitCode()

    def isUp(self):
        return self.envRunner.isUp()

    def isHealthy(self):
        return self.envRunner.isHealthy()

    def skip(self):
        raise unittest.SkipTest()

    def skipOnDebugger(self):
        if self.isDebugger():
            self.skip()

    def skipOnCluster(self):
        if self.isCluster():
            self.skip()

    def skipOnAOF(self):
        if self.useAof:
            self.skip()

    def skipOnSlave(self):
        if self.useSlaves:
            self.skip()

    def skipOnVersionSmaller(self, _version):
        res = self.con.execute_command('INFO')
        if(version.parse(res['redis_version']) < version.parse(_version)):
            self.skip() # copy exists only from version 6

    def isUnixSocket(self):
        return self.envRunner.isUnixSocket()

    def isTcp(self):
        return self.envRunner.isTcp()

    def skipOnTcp(self):
        if self.isTcp():
            self.skip()

    def skipOnUnixSocket(self):
        if self.isUnixSocket():
            self.skip()

    def skipOnEnterpriseCluster(self):
        if self.isEnterpiseCluster():
            self.skip()

    _mm = {
        'assertEquals': assertEqual,
        'assertListEqual': assertEqual,
        'retry_with_reload': reloadingIterator,
        'retry_with_rdb_reload': reloadingIterator,
        'reloading_iterator': reloadingIterator,
        'dump_and_reload': dumpAndReload,
        'restart_and_reload': restartAndReload,
        'execute_command': cmd,
        'assertIn': assertContains,
        'assertNotIn': assertNotContains,
        'is_cluster': isCluster,
        'is_enterprise_redis_clusterEnv':isEnterpiseCluster
    }
    for k, v in _mm.items():
        locals().update({k:genDeprecated(k, v)})



import platform
import os.path


class Valgrind(object):
    is_interactive = False

    def __init__(self, options, suppressions=None, fail_on_errors=True, leakcheck=True ):
        self.options = options
        self.suppressions = suppressions
        self.leakcheck = leakcheck
        self.fail_on_errors = fail_on_errors

    def generate_command(self, logfile=None):
        cmd = ['valgrind']
        if self.fail_on_errors == True:
            cmd += ['--error-exitcode=255']
        for option in self.options.split():
            cmd += [option]
        if self.leakcheck:
            if '--leak-check=full' not in self.options:
                    cmd += ['--leak-check=full']
            if '--errors-for-leak-kinds=definite' not in self.options:
                    cmd += ['--errors-for-leak-kinds=definite']
        if self.suppressions:
            cmd += ['--suppressions=' + self.suppressions]
        if logfile:
            cmd += ['--log-file=' + logfile]
        return cmd


class GenericInteractiveDebugger(object):
    is_interactive = True

    def __init__(self, cmdline):
        self.args = cmdline.split()

    def generate_command(self, *argc, **kw):
        return [self.args]


class GDB(GenericInteractiveDebugger):
    def __init__(self, cmdline="gdb"):
        super(GDB, self).__init__(cmdline)

    def generate_command(self, *argc, **kw):
        return ['gdb', '-ex', 'run', '--args']


class CGDB(GenericInteractiveDebugger):
    def __init__(self, cmdline="cgdb"):
        super(CGDB, self).__init__(cmdline)

    def generate_command(self, *argc, **kw):
        return ['cgdb', '-ex', 'run', '--args']


class LLDB(GenericInteractiveDebugger):
    def __init__(self, cmdline="lldb"):
        super(LLDB, self).__init__(cmdline)

    def generate_command(self, *argc, **kw):
        return ['lldb', '-o', 'run', '--']


if platform.system() == 'Darwin':
    DefaultInteractiveDebugger = LLDB
else:
    DefaultInteractiveDebugger = GDB

default_interactive_debugger = DefaultInteractiveDebugger()

def set_interactive_debugger(debugger):
    global default_interactive_debugger
    cmd = os.path.basename(debugger.split()[0])
    if cmd == 'gdb':
        default_interactive_debugger = GDB(debugger)
    elif cmd == 'cgdb':
        default_interactive_debugger = CGDB(debugger)
    elif cmd == 'lldb':
        default_interactive_debugger = LLDB(debugger)




# This attribute is the only one place that the version number is written down,
# so there is only one place to change it when the version number changes.
import pkg_resources
try:
    __version__ = pkg_resources.get_distribution('RLTest').version
except (pkg_resources.DistributionNotFound, AttributeError):
    __version__ = "99.99.99"  # like redis modules



from __future__ import print_function

import os
import subprocess
import sys
import time
import uuid
import platform
import psutil

import redis

from .random_port import get_random_port
from .utils import Colors, wait_for_conn, fix_modules, fix_modulesArgs

MASTER = 'master'
SLAVE = 'slave'


class StandardEnv(object):
    def __init__(self, redisBinaryPath, port=6379, modulePath=None, moduleArgs=None, outputFilesFormat=None,
                 dbDirPath=None, useSlaves=False, serverId=1, password=None, libPath=None, clusterEnabled=False, decodeResponses=False,
                 useAof=False, useRdbPreamble=True, debugger=None, sanitizer=None, noCatch=False, noLog=False, unix=False, verbose=False, useTLS=False,
                 tlsCertFile=None, tlsKeyFile=None, tlsCaCertFile=None, clusterNodeTimeout=None, tlsPassphrase=None, enableDebugCommand=False, protocol=2,
                 terminateRetries=None, terminateRetrySecs=None, enableProtectedConfigs=False, loglevel=None):
        self.uuid = uuid.uuid4().hex
        self.redisBinaryPath = os.path.expanduser(redisBinaryPath) if redisBinaryPath.startswith(
            '~/') else redisBinaryPath

        self.modulePath = fix_modules(modulePath)
        self.moduleArgs = fix_modulesArgs(self.modulePath, moduleArgs, haveSeqs=False)
        self.outputFilesFormat = self.uuid + '.' + outputFilesFormat
        self.useSlaves = useSlaves
        self.masterServerId = serverId
        self.password = password
        self.clusterEnabled = clusterEnabled
        self.decodeResponses = decodeResponses
        self.useAof = useAof
        self.useRdbPreamble = useRdbPreamble
        self.envIsUp = False
        self.debugger = debugger
        self.sanitizer = sanitizer
        self.noCatch = noCatch
        self.noLog = noLog
        self.loglevel = loglevel
        self.environ = os.environ.copy()
        self.useUnix = unix
        self.dbDirPath = dbDirPath
        self.masterProcess = None
        self.masterStdout = None
        self.masterStderr = None
        self.masterExitCode = None
        self.slaveProcess = None
        self.slaveStdout = None
        self.slaveStderr = None
        self.slaveExitCode = None
        self.verbose = verbose
        self.role = MASTER
        self.useTLS = useTLS
        self.tlsCertFile = tlsCertFile
        self.tlsKeyFile = tlsKeyFile
        self.tlsCaCertFile = tlsCaCertFile
        self.clusterNodeTimeout = clusterNodeTimeout
        self.tlsPassphrase = tlsPassphrase
        self.enableDebugCommand = enableDebugCommand
        self.enableProtectedConfigs = enableProtectedConfigs
        self.protocol = protocol
        self.terminateRetries = terminateRetries
        self.terminateRetrySecs = terminateRetrySecs

        if port > 0:
            self.port = port
            self.slavePort = port + 1 if self.useSlaves else 0
        elif port == 0:
            self.port = get_random_port()
            self.slavePort = get_random_port() if self.useSlaves else 0
        else:
            self.port = -1
            self.slavePort = -1

        if self.useUnix:
            if self.clusterEnabled:
                raise ValueError('Unix sockets cannot be used with cluster mode')
            self.port = -1

        if self.useTLS:
            if self.useUnix:
                raise ValueError('Unix sockets cannot be used with TLS enabled mode')
            if self.tlsCertFile is None:
                raise ValueError('When useTLS option is True tlsCertFile must be defined')
            if os.path.isfile(self.tlsCertFile) is False:
                raise ValueError(
                    'When useTLS option is True tlsCertFile must exist. specified path {}'.format(self.tlsCertFile))
            if self.tlsKeyFile is None:
                raise ValueError('When useTLS option is True tlsKeyFile must be defined')
            if os.path.isfile(self.tlsKeyFile) is False:
                raise ValueError(
                    'When useTLS option is True tlsKeyFile must exist. specified path {}'.format(self.tlsKeyFile))
            if self.tlsCaCertFile is None:
                raise ValueError('When useTLS option is True tlsCaCertFile must be defined')
            if not os.path.isfile(self.tlsCaCertFile):
                raise ValueError(
                    'When useTLS option is True tlsCaCertFile must exist. specified path {}'.format(self.tlsCaCertFile))

        if libPath:
            self.libPath = os.path.expanduser(libPath) if libPath.startswith('~/') else libPath
        else:
            self.libPath = None
        if self.libPath:
            if 'LD_LIBRARY_PATH' in self.environ.keys():
                self.environ['LD_LIBRARY_PATH'] = self.libPath + ":" + self.environ['LD_LIBRARY_PATH']
            else:
                self.environ['LD_LIBRARY_PATH'] = self.libPath

        self.masterCmdArgs = self.createCmdArgs(MASTER)
        self.masterOSEnv = self.createCmdOSEnv(MASTER)
        if self.useSlaves:
            self.slaveServerId = serverId + 1
            self.slaveCmdArgs = self.createCmdArgs(SLAVE)
            self.slaveOSEnv = self.createCmdOSEnv(SLAVE)

        self.envIsHealthy = True

    def _getFileName(self, role, suffix):
        return (self.outputFilesFormat + suffix) % (
            'master-%d' % self.masterServerId if role == MASTER else 'slave-%d' % self.slaveServerId)

    def _getValgrindFilePath(self, role):
        return os.path.join(self.dbDirPath, self._getFileName(role, '.valgrind.log'))

    def getMasterPort(self):
        return self.port

    def getPassword(self):
        return self.password

    def getUnixPath(self, role):
        basename = '{}-{}.sock'.format(self.uuid, role)
        return os.path.abspath(os.path.join(self.dbDirPath, basename))

    def getTLSCertFile(self):
        return os.path.abspath(self.tlsCertFile)

    def getTLSKeyFile(self):
        return os.path.abspath(self.tlsKeyFile)

    def getTLSCACertFile(self):
        return os.path.abspath(self.tlsCaCertFile)

    @property
    def has_interactive_debugger(self):
        return self.debugger and self.debugger.is_interactive

    def _getRedisVersion(self):
        options = {
            'stderr': subprocess.PIPE,
            'stdin': subprocess.PIPE,
            'stdout': subprocess.PIPE,
        }
        p = subprocess.Popen(args=[self.redisBinaryPath, '--version'], **options)
        while p.poll() is None:
            time.sleep(0.1)
        exit_code = p.poll()
        if exit_code != 0:
            raise Exception('Could not extract Redis version')
        out, err = p.communicate()
        out = out.decode('utf-8')
        v = out[out.find("v=") + 2:out.find("sha=") - 1].split('.')
        return int(v[0]) * 10000 + int(v[1]) * 100 + int(v[2])

    def createCmdArgs(self, role):
        cmdArgs = []
        if self.debugger and role == MASTER and self.masterServerId == 1:
            cmdArgs += self.debugger.generate_command(self._getValgrindFilePath(role) if not self.noCatch else None)

        cmdArgs += [self.redisBinaryPath]

        if self.port > -1:
            if self.useTLS:
                cmdArgs += ['--port', str(0), '--tls-port', str(self.getPort(role))]
            else:
                cmdArgs += ['--port', str(self.getPort(role))]
        else:
            cmdArgs += ['--port', str(0), '--unixsocket', self.getUnixPath(role)]

        if self.modulePath:
            if self.moduleArgs and len(self.modulePath) != len(self.moduleArgs):
                print(Colors.Bred('Number of module args sets in Env does not match number of modules'))
                print(self.modulePath)
                print(self.moduleArgs)
                sys.exit(1)
            for pos, module in enumerate(self.modulePath):
                cmdArgs += ['--loadmodule', module]
                if self.moduleArgs:
                    module_args = self.moduleArgs[pos]
                    if module_args:
                        # make sure there are no spaces within args
                        args = []
                        for arg in module_args:
                            if arg.strip() != '':
                                args += arg.split(' ')
                        cmdArgs += args

        if self.dbDirPath is not None:
            cmdArgs += ['--dir', self.dbDirPath]
        if self.noLog:
            cmdArgs += ['--logfile', '/dev/null']
        elif self.outputFilesFormat is not None and not self.noCatch:
            cmdArgs += ['--logfile', self._getFileName(role, '.log')]
        if self.loglevel is not None:
            cmdArgs += ['--loglevel', self.loglevel]
        if self.outputFilesFormat is not None:
            cmdArgs += ['--dbfilename', self._getFileName(role, '.rdb')]
        if role == SLAVE:
            cmdArgs += ['--slaveof', 'localhost', str(self.port)]
            if self.password:
                cmdArgs += ['--masterauth', self.password]
        if self.password:
            cmdArgs += ['--requirepass', self.password]
        if self.clusterEnabled and role is not SLAVE:
            # creating .cluster.conf in /tmp as lock fails on NFS
            cmdArgs += ['--cluster-enabled', 'yes', '--cluster-config-file', '/tmp/' + self._getFileName(role, '.cluster.conf'),
                        '--cluster-node-timeout', '5000' if self.clusterNodeTimeout is None else str(self.clusterNodeTimeout)]
            if self.useTLS:
                cmdArgs += ['--tls-cluster', 'yes']
        if self.useAof:
            cmdArgs += ['--appendonly', 'yes']
            cmdArgs += ['--appendfilename', self._getFileName(role, '.aof')]
            if not self.useRdbPreamble:
                cmdArgs += ['--aof-use-rdb-preamble', 'no']
        if self.useTLS:
            cmdArgs += ['--tls-cert-file', self.getTLSCertFile()]
            cmdArgs += ['--tls-key-file', self.getTLSKeyFile()]
            cmdArgs += ['--tls-ca-cert-file', self.getTLSCACertFile()]
            if self.tlsPassphrase:
                cmdArgs += ['--tls-key-file-pass', self.tlsPassphrase]

            cmdArgs += ['--tls-replication', 'yes']

        if self._getRedisVersion() > 70000:
            if self.enableDebugCommand:
                cmdArgs += ['--enable-debug-command', 'yes']
            if self.enableProtectedConfigs:
                cmdArgs += ['--enable-protected-configs', 'yes']

        return cmdArgs

    def createCmdOSEnv(self, role):
        if self.sanitizer != 'addr' and self.sanitizer != 'address':
            return self.environ
        osenv = self.environ.copy()
        san_log = self._getFileName(role, '.asan.log')
        asan_options = osenv.get("ASAN_OPTIONS")
        osenv["ASAN_OPTIONS"] = "{OPT}:log_path={DIR}".format(OPT=asan_options, DIR=san_log)
        return osenv

    def waitForRedisToStart(self, con, proc):
        wait_for_conn(con, proc, retries=1000 if self.debugger else 200)
        self._waitForAOFChild(con)

    def getPid(self, role):
        return self.masterProcess.pid if role == MASTER else self.slaveProcess.pid

    def getPort(self, role):
        return self.port if role == MASTER else self.slavePort

    def getServerId(self, role):
        return self.masterServerId if role == MASTER else self.slaveServerId

    def _printEnvData(self, prefix='', role=MASTER):
        print(Colors.Yellow(prefix + 'pid: %d' % (self.getPid(role))))
        if self.useUnix:
            print(Colors.Yellow(prefix + 'unix_socket_path: %s' % (self.getUnixPath(role))))
        else:
            print(Colors.Yellow(prefix + 'port: %d' % (self.getPort(role))))
        print(Colors.Yellow(prefix + 'binary path: %s' % (self.redisBinaryPath)))
        print(Colors.Yellow(prefix + 'server id: %d' % (self.getServerId(role))))
        print(Colors.Yellow(prefix + 'using debugger: {}'.format(bool(self.debugger))))
        if self.modulePath:
            print(Colors.Yellow(prefix + 'module: %s' % (self.modulePath)))
            if self.moduleArgs:
                print(Colors.Yellow(prefix + 'module args: %s' % (self.moduleArgs)))
        if self.outputFilesFormat:
            print(Colors.Yellow(prefix + 'log file: %s' % (self._getFileName(role, '.log'))))
            print(Colors.Yellow(prefix + 'db file name: %s' % self._getFileName(role, '.rdb')))
        if self.dbDirPath:
            print(Colors.Yellow(prefix + 'db dir path: %s' % (self.dbDirPath)))
        if self.libPath:
            print(Colors.Yellow(prefix + 'library path: %s' % (self.libPath)))
        if self.useTLS:
            print(Colors.Yellow(prefix + 'TLS Cert File: %s' % (self.getTLSCertFile())))
            print(Colors.Yellow(prefix + 'TLS Key File: %s' % (self.getTLSKeyFile())))
            print(Colors.Yellow(prefix + 'TLS CA Cert File: %s' % (self.getTLSCACertFile())))

    def printEnvData(self, prefix=''):
        print(Colors.Yellow(prefix + 'master:'))
        self._printEnvData(prefix + '\t', MASTER)
        if self.useSlaves:
            print(Colors.Yellow(prefix + 'slave:'))
            self._printEnvData(prefix + '\t', SLAVE)

    def getInformationBeforeDispose(self):
        res = {}
        instances = [(MASTER, self.getConnection(), self.masterProcess)]
        if self.useSlaves:
            instances.append((SLAVE, self.getSlaveConnection(), self.slaveProcess))
        for role, conn, proc in instances:
            logs = None
            info = None
            try:
                with open(os.path.join(self.dbDirPath, self._getFileName(role, '.log'))) as f:
                    logs = f.read()
            except os.FileNoteFoundError:
                pass
            try:
                info = conn.execute_command('info', 'everything')
            except redis.exceptions.RedisError:
                pass
            res[role] = {
                'info': info,
                'logs': logs,
            }
        return res

    def getInformationAfterDispose(self):
        res = {}
        instances = [(MASTER, self.masterStdout, self.masterStderr)]
        if self.useSlaves:
            instances.append((SLAVE, self.slaveStdout, self.slaveStderr))
        for role, stdout, stderr in instances:
            stdoutStr = None
            stderrStr = None
            try:
                stdoutStr = stdout.read().decode('utf8')
            except (NameError, AttributeError):
                pass

            try:
                stderrStr = stderr.read().decode('utf8')
            except (NameError, AttributeError):
                pass

            res[role] = {
                'stdout': stdoutStr,
                'stderr': stderrStr,
            }
        return res

    def startEnv(self, masters = True, slaves = True):
        if self.envIsUp and self.envIsHealthy:
            return  # env is already up
        stdoutPipe = subprocess.PIPE
        stderrPipe = subprocess.PIPE
        stdinPipe = subprocess.PIPE
        if self.noCatch:
            stdoutPipe = sys.stdout
            stderrPipe = sys.stderr

        if self.has_interactive_debugger:
            stdinPipe = sys.stdin

        options = {
            'stderr': stderrPipe,
            'stdin': stdinPipe,
            'stdout': stdoutPipe,
        }

        if self.verbose:
            print(Colors.Green("Redis master command: " + ' '.join(self.masterCmdArgs)))
        if masters and self.masterProcess is None:
            self.masterProcess = subprocess.Popen(args=self.masterCmdArgs, env=self.masterOSEnv, **options)
            time.sleep(0.1)
            if self._isAlive(self.masterProcess):
                con = self.getConnection()
                self.waitForRedisToStart(con, self.masterProcess)
            else:
                self.masterProcess = None
        if self.useSlaves and slaves and self.slaveProcess is None:
            if self.verbose:
                print(Colors.Green("Redis slave command: " + ' '.join(self.slaveCmdArgs)))
            self.slaveProcess = subprocess.Popen(args=self.slaveCmdArgs, env=self.slaveOSEnv, **options)
            time.sleep(0.1)
            if self._isAlive(self.slaveProcess):
                con = self.getSlaveConnection()
                self.waitForRedisToStart(con, self.slaveProcess)
            else:
                self.slaveProcess = None

        self.envIsUp = self.masterProcess is not None or self.slaveProcess is not None
        self.envIsHealthy = self.masterProcess is not None and (self.slaveProcess is not None if self.useSlaves else True)

        # self.masterStdout = self.masterProcess.stdout if self.masterProcess else None
        # self.masterStderr = self.masterProcess.stderr if self.masterProcess else None

        # if self.slaveProcess is not None:
        #     self.slaveStdout = self.slaveProcess.stdout if self.slaveProcess else None
        #     self.slaveStderr = self.slaveProcess.stderr if self.slaveProcess else None

    def _isAlive(self, process):
        if not process:
            return False
        # check if child process has terminated
        if process.poll() is None:
            return True
        return False

    def _stopProcess(self, role):
        process = self.masterProcess if role == MASTER else self.slaveProcess
        serverId = self.masterServerId if role == MASTER else self.slaveServerId
        if not self._isAlive(process):
            if not self.has_interactive_debugger:
                # on interactive debugger its expected that then process will not be alive
                print('\t' + Colors.Bred('process is not alive, might have crash durring test execution, '
                                         'check this out. server id : %s' % str(serverId)))
                if self.outputFilesFormat is not None and not self.noCatch:
                    self.verbose_analyse_server_log(role)
            return
        try:
            if platform.system() == 'Darwin':
                # On macOS, with lldb, killing lldb process does not terminate inferior processes
                p0 = psutil.Process(pid=process.pid)
                pchi = p0.children(recursive=True)
                for p in pchi:
                    try:
                        p.terminate()
                        p.wait()
                    except:
                        pass

            if self.terminateRetries is None:
                # ask once, then wait for process to exit
                process.terminate()
                while True:
                    if process.poll() is None:  # None returns if the processes is not finished yet, retry until redis exits
                        time.sleep(0.1)
                    else:
                        break
            else:
                # keep asking every few seconds until process has exited, otherwise kill
                if self.terminateRetrySecs is None:
                    self.terminateRetrySecs = 1
                done = False
                for i in range(0, self.terminateRetries):
                    process.terminate()
                    if process.poll() is None:  # None returns if the processes is not finished yet, retry until redis exits
                        time.sleep(self.terminateRetrySecs)
                    else:
                        done = True
                        break
                if not done:
                    process.kill()

            if role == MASTER:
                self.masterExitCode = process.poll()
            else:
                self.slaveExitCode = process.poll()
        except OSError as e:
            print('\t' + Colors.Bred(
                'OSError caught while waiting for {0} process to end: {1}'.format(role, e.__str__())))
            pass

    def verbose_analyse_server_log(self, role):
        path = "{0}".format(self._getFileName(role, '.log'))
        if self.dbDirPath is not None:
            path = "{0}/{1}".format(self.dbDirPath, self._getFileName(role, '.log'))
        print('\t' + Colors.Bred('check the redis log at: {0}'.format(path)))
        print('\t' + Colors.Yellow('Printing only REDIS BUG REPORT START and STACK TRACE'))
        with open(path) as file:
            bug_report_found = False
            for line in file:
                if "REDIS BUG REPORT START" in line:
                    bug_report_found = True
                if "------ INFO OUTPUT ------" in line:
                    break
                if bug_report_found is True:
                    print('\t\t' + Colors.Yellow(line.rstrip()))

    def stopEnv(self, masters = True, slaves = True):
        if self.masterProcess is not None and masters is True:
            self._stopProcess(MASTER)
            self.masterProcess = None
        if self.useSlaves and self.slaveProcess is not None and slaves is True:
            self._stopProcess(SLAVE)
            self.slaveProcess = None
        self.envIsUp = self.masterProcess is not None or self.slaveProcess is not None
        self.envIsHealthy = self.masterProcess is not None and (self.slaveProcess is not None if self.useSlaves else True)


    def _getConnection(self, role):
        if self.useUnix:
            return redis.StrictRedis(unix_socket_path=self.getUnixPath(role),
                                     password=self.password, decode_responses=self.decodeResponses, protocol=self.protocol)
        elif self.useTLS:
            return redis.StrictRedis('localhost', self.getPort(role),
                                     password=self.password,
                                     ssl=True,
                                     ssl_password=self.tlsPassphrase,
                                     ssl_keyfile=self.getTLSKeyFile(),
                                     ssl_certfile=self.getTLSCertFile(),
                                     ssl_cert_reqs=None,
                                     ssl_ca_certs=self.getTLSCACertFile(),
                                     decode_responses=self.decodeResponses,
                                     protocol=self.protocol
                                     )
        else:
            return redis.StrictRedis('localhost', self.getPort(role),
                                     password=self.password, decode_responses=self.decodeResponses, protocol=self.protocol)

    def getConnection(self, shardId=1):
        return self._getConnection(MASTER)

    # List containing a connection for each of the master nodes
    def getOSSMasterNodesConnectionList(self):
        return [self.getConnection()]

    def getSlaveConnection(self):
        if self.useSlaves:
            return self._getConnection(SLAVE)
        raise Exception('asked for slave connection but no slave exists')

    # List of nodes that initial bootstrapping can be done from
    def getMasterNodesList(self):
        node_info = {"host": None, "port": None, "unix_socket_path": None, "password": None}
        node_info["password"] = self.password
        if self.useUnix:
            node_info["unix_socket_path"] = self.getUnixPath(MASTER)

        else:
            node_info["host"] = 'localhost'
            node_info["port"] = self.getPort(MASTER)
        return [node_info]

    def getConnectionByKey(self, key, command):
        return self.getConnection()

    def flush(self):
        self.getConnection().flushall()

    def _waitForAOFChild(self, con):
        import time
        # Wait until file is available
        while True:
            info = con.info('persistence')
            if info['aof_rewrite_scheduled'] or info['aof_rewrite_in_progress']:
                time.sleep(0.1)
            else:
                break

    def dumpAndReload(self, restart=False, shardId=None, timeout_sec=0):
        conns = []
        conns.append(self.getConnection())
        if self.useSlaves:
            conns.append(self.getSlaveConnection())
        if restart:
            for con in conns:
                self._waitForAOFChild(con)
                con.bgrewriteaof()
                self._waitForAOFChild(con)

            self.stopEnv()
            self.startEnv()
        else:
            [con.save() for con in conns]
            try:
                # Given we've already saved on the prior step, there is no need to SAVE again on the DEBUG RELOAD
                [con.execute_command('DEBUG', 'RELOAD', 'NOSAVE') for con in conns]
            except redis.RedisError as err:
                raise err

    def broadcast(self, *cmd):
        try:
            self.getConnection().execute_command(*cmd)
        except Exception as e:
            print(e)

    def checkExitCode(self):
        ret = True

        if self.masterExitCode != 0:
            print('\t' + Colors.Bred('bad exit code for serverId %s' % str(self.masterServerId)))
            ret = False
        if self.useSlaves and (self.slaveExitCode is None or self.slaveExitCode != 0):
            print('\t' + Colors.Bred('bad exit code for serverId %s' % str(self.slaveServerId)))
            ret = False
        return ret

    def isUp(self):
        return self.envIsUp

    def isHealthy(self):
        return self.envIsHealthy

    def isUnixSocket(self):
        return self.useUnix

    def isTcp(self):
        return not (self.useUnix)

    def isTLS(self):
        return self.useTLS

    def exists(self, val):
        return self.getConnection().exists(val)

    def hmset(self, *args):
        return self.getConnection().hmset(*args)

    def keys(self, reg):
        return self.getConnection().keys(reg)

    def setTerminateRetries(self, retries=3, seconds=1):
        self.terminateRetries = retries
        self.terminateRetrySecs = seconds



from __future__ import print_function

import argparse
import io
import os
import cmd
import traceback
import sys
import shutil
import inspect
import unittest
import time
import shlex
import json
from multiprocessing import Process, Queue

from RLTest.env import Env, TestAssertionFailure, Defaults
from RLTest.utils import Colors, fix_modules, fix_modulesArgs
from RLTest.loader import TestLoader
from RLTest.Enterprise import binaryrepo
from RLTest import debuggers
from RLTest._version import __version__

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

RLTest_CONFIG_FILE_PREFIX = '@'
RLTest_CONFIG_FILE_NAME = 'config.txt'

class CustomArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwrags):
        super(CustomArgumentParser, self).__init__(*args, **kwrags)

    def convert_arg_line_to_args(self, line):
        for arg in shlex.split(line):
            if not arg.strip():
                continue
            if arg[0] == '#':
                break
            yield arg


class MyCmd(cmd.Cmd):

    def __init__(self, env):
        cmd.Cmd.__init__(self)
        self.env = env
        self.prompt = '> '
        try:
            commands_reply = env.cmd('command')
        except Exception:
            return
        commands = [c[0] for c in commands_reply]
        for c in commands:
            if type(c)==bytes:
                c=c.decode('utf-8')
            setattr(MyCmd, 'do_' + c, self._create_functio(c))

    def _exec(self, command):
        self.env.expect(*command).prettyPrint()

    def _create_functio(self, command):
        c = command
        return lambda self, x: self._exec([c] + shlex.split(x))

    def do_exec(self, line):
        self.env.expect(*shlex.split(line)).prettyPrint()

    def do_print(self, line):
        '''
        print
        '''
        print('print')

    def do_stop(self, line):
        '''
        print
        '''
        print('BYE BYE')
        return True

    def do_cluster_conn(self, line):
        '''
        move to oss-cluster connection
        '''
        if self.env.env == 'oss-cluster':
            self.env.con = self.env.envRunner.getClusterConnection()
            print('moved to cluster connection')
        else:
            print('cluster connection only available on oss-cluster env')

    def do_normal_conn(self, line):
        '''
        move to normal connection (will connect to the first shard on oss-cluster)
        '''
        self.env.con = self.env.envRunner.getConnection()
        print('moved to normal connection (first shard on oss-cluster)')

    do_exit = do_stop


parser = CustomArgumentParser(fromfile_prefix_chars=RLTest_CONFIG_FILE_PREFIX,
                              formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                              description='Test Framework for redis and redis module')
parser.add_argument(
    '--version', action='store_const', const=True, default=False,
    help='Print RLTest version and exit')

parser.add_argument(
    '--module', default=None, action='append',
    help='path to the module file. '
         'You can use `--module` more than once but it imples that you explicitly specify `--module-args` as well. '
         'Notice that on enterprise the file should be a zip file packed with [RAMP](https://github.com/RedisLabs/RAMP).')

parser.add_argument(
    '--module-args', default=None, action='append', nargs='*',
    help='arguments to give to the module on loading')

parser.add_argument(
    '--env', '-e', default='oss', choices=['oss', 'oss-cluster', 'enterprise', 'enterprise-cluster', 'existing-env', 'cluster_existing-env'],
    help='env on which to run the test')

parser.add_argument(
    '-p', '--redis-port', type=int, default=6379,
    help='Redis server port')

parser.add_argument(
    '--existing-env-addr', default='localhost:6379',
    help='Address of existing env, relevent only when running with existing-env, cluster_existing-env')

parser.add_argument(
    '--shards_ports',
    help=' list of ports, the shards are listening to, relevent only when running with cluster_existing-env')

parser.add_argument(
    '--cluster_address',
    help='enterprise cluster ip, relevent only when running with cluster_existing-env')

parser.add_argument(
    '--oss_password', default=None,
    help='set redis password, relevant for oss and oss-cluster environment')

parser.add_argument(
    '--cluster_node_timeout', default=5000,
    help='sets the node timeout on cluster in milliseconds')

parser.add_argument(
    '--cluster_credentials',
    help='enterprise cluster cluster_credentials "username:password", relevent only when running with cluster_existing-env')

parser.add_argument(
    '--internal_password', default='',
    help='Give an ability to execute commands on shards directly, relevent only when running with cluster_existing-env')

parser.add_argument(
    '--oss-redis-path', default='redis-server',
    help='path to the oss redis binary')

parser.add_argument(
    '--enterprise-redis-path', default=os.path.join(binaryrepo.REPO_ROOT, 'opt/redislabs/bin/redis-server'),
    help='path to the entrprise redis binary')

parser.add_argument(
    '--stop-on-failure', action='store_const', const=True, default=False,
    help='stop running on failure')

parser.add_argument(
    '-x', '--exit-on-failure', action='store_true',
    help='Stop test execution and exit on first assertion failure')

parser.add_argument(
    '--verbose', '-v', action='count', default=0,
    help='print more information about the test')

parser.add_argument(
    '--debug', action='store_const', const=True, default=False,
    help='stop before each test allow gdb attachment')

parser.add_argument(
    '-t', '--test', metavar='TEST', action='append', help='test to run, in the form of "file:test"')

parser.add_argument(
    '-f', '--tests-file', metavar='FILE', action='append', help='file containing test to run, in the form of "file:test"')

parser.add_argument(
    '-F', '--failed-tests-file', metavar='FILE', help='destination file for failed tests')

parser.add_argument(
    '--env-only', action='store_const', const=True, default=False,
    help='start the env but do not run any tests')

parser.add_argument(
    '--clear-logs', action='store_const', const=True, default=False,
    help='deleting the log direcotry before the execution')

parser.add_argument(
    '--log-dir', default='./logs',
    help='directory to write logs to')

parser.add_argument(
    '--log-level', default=None, metavar='LEVEL', choices=['debug', 'verbose', 'notice', 'warning'],
    help='sets the server log level')

parser.add_argument(
    '--use-slaves', action='store_const', const=True, default=False,
    help='run env with slaves enabled')

parser.add_argument(
    '--shards-count', default=1, type=int,
    help='Number shards in bdb')

parser.add_argument(
    '--download-enterprise-binaries', action='store_const', const=True, default=False,
    help='run env with slaves enabled')

parser.add_argument(
    '--proxy-binary-path', default=os.path.join(binaryrepo.REPO_ROOT, 'opt/redislabs/bin/dmcproxy'),
    help='dmc proxy binary path')

parser.add_argument(
    '--enterprise-lib-path', default=os.path.join(binaryrepo.REPO_ROOT, 'opt/redislabs/lib/'),
    help='path of needed libraries to run enterprise binaries')

parser.add_argument(
    '-r', '--env-reuse', action='store_const', const=True, default=False,
    help='reuse exists env, this feature is based on best efforts, if the env can not be reused then it will be taken down.')

parser.add_argument(
    '--use-aof', action='store_const', const=True, default=False,
    help='use aof instead of rdb')

parser.add_argument(
    '--use-rdb-preamble', action='store_const', const=True, default=True,
    help='use rdb preamble when rewriting aof file')

parser.add_argument(
    '--debug-print', action='store_const', const=True, default=False,
    help='print debug messages')

parser.add_argument(
    '-V', '--vg', '--use-valgrind', action='store_const', const=True, default=False,
    dest='use_valgrind',
    help='running redis under valgrind (assuming valgrind is install on the machine)')

parser.add_argument(
    '--vg-suppressions', default=None, help='path valgrind suppressions file')
parser.add_argument(
    '--vg-options', default=None, dest='vg_options', help='valgrind [options]')
parser.add_argument(
    '--vg-no-leakcheck', action='store_true', help="Don't perform a leak check")
parser.add_argument(
    '--vg-verbose', action='store_true', help="Don't log valgrind output. "
                                              "Output to screen directly")
parser.add_argument(
    '--vg-no-fail-on-errors', action='store_true', dest='vg_no_fail_on_errors', help="Dont Fail test when valgrind reported any errors in the run."
                                                  "By default on RLTest the return value from Valgrind will be used to fail the tests."
                                                  "Use this option when you wish to dry-run valgrind but not fail the test on valgrind reported errors."
)

parser.add_argument(
    '--sanitizer', default=None, help='type of CLang sanitizer (addr|mem)')

parser.add_argument(
    '-i', '--interactive-debugger', action='store_const', const=True, default=False,
    help='runs the redis on a debuger (gdb/lldb) interactivly.'
         'debugger interactive mode is only possible on a single process and so unsupported on cluste or with slaves.'
         'it is also not possible to use valgrind on interactive mode.'
         'interactive mode direcly applies: --no-output-catch and --stop-on-failure.'
         'it is also implies that only one test will be run (if --env-only was not specify), an error will be raise otherwise.')

parser.add_argument('--debugger', help='Run specified command line as the debugger')

parser.add_argument(
    '-s', '--no-output-catch', action='store_const', const=True, default=False,
    help='all output will be written to the stdout, no log files.')

parser.add_argument(
    '--verbose-information-on-failure', action='store_const', const=True, default=False,
    help='Print a verbose information on test failure')

parser.add_argument(
    '--enable-debug-command', action='store_const', const=True, default=False,
    help='On Redis 7, debug command need to be enabled in order to be used.')

parser.add_argument(
    '--enable-protected-configs', action='store_const', const=True, default=False,
    help='On Redis 7, this option needs to be enabled in order to change protected configuration in runtime.')

parser.add_argument('--check-exitcode', help='Check redis process exit code',
                    default=False, action='store_true')

parser.add_argument('--unix', help='Use Unix domain sockets instead of TCP',
                    default=False, action='store_true')

parser.add_argument('--randomize-ports',
                    help='Randomize Redis listening port assignment rather than'
                    'using default port',
                    default=False, action='store_true')

parser.add_argument('--parallelism', help='Run tests in parallel', default=1, type=int)

parser.add_argument(
    '--collect-only', action='store_true',
    help='Collect the tests and exit')

parser.add_argument('--tls', help='Enable TLS Support and disable the non-TLS port completely. TLS connections will be available at the default non-TLS ports.',
                    default=False, action='store_true')

parser.add_argument(
    '--tls-cert-file', default=None, help='/path/to/redis.crt')

parser.add_argument(
    '--tls-key-file', default=None, help='/path/to/redis.key')

parser.add_argument(
    '--tls-ca-cert-file', default=None, help='/path/to/ca.crt')

parser.add_argument(
    '--tls-passphrase', default=None, help='passphrase to use on decript key file')

class EnvScopeGuard:
    def __init__(self, runner):
        self.runner = runner

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        self.runner.takeEnvDown()


class RLTest:
    def __init__(self):
        # adding the current path to sys.path for test import puspused
        sys.path.append(os.getcwd())

        configFilePath = './%s' % RLTest_CONFIG_FILE_NAME
        if os.path.exists(configFilePath):
            args = ['%s%s' % (RLTest_CONFIG_FILE_PREFIX, RLTest_CONFIG_FILE_NAME)] + sys.argv[1:]
        else:
            args = sys.argv[1:]
        self.args = parser.parse_args(args=args)

        if self.args.version:
            print(Colors.Green('RLTest version {}'.format(__version__)))
            sys.exit(0)

        if self.args.redis_port not in range(1, pow(2, 16)):
            print(Colors.Bred(f'requested port {self.args.redis_port} is not valid'))
            sys.exit(1)

        if self.args.interactive_debugger:
            if self.args.env != 'oss' and not (self.args.env == 'oss-cluster' and Defaults.num_shards == 1) and self.args.env != 'enterprise':
                print(Colors.Bred('interactive debugger can only be used on non cluster env'))
                sys.exit(1)
            if self.args.use_valgrind:
                print(Colors.Bred('can not use valgrind with interactive debugger'))
                sys.exit(1)
            if self.args.use_slaves:
                print(Colors.Bred('can not use slaves with interactive debugger'))
                sys.exit(1)

            self.args.no_output_catch = True
            self.args.stop_on_failure = True

        if self.args.download_enterprise_binaries:
            br = binaryrepo.BinaryRepository()
            br.download_binaries()

        if self.args.clear_logs:
            if os.path.exists(self.args.log_dir):
                try:
                    shutil.rmtree(self.args.log_dir)
                except Exception as e:
                    print(e, file=sys.stderr)

        debugger = None
        if self.args.debugger:
            if self.args.env.endswith('existing-env'):
                print(Colors.Bred('can not use debug with existing-env'))
                sys.exit(1)
            debuggers.set_interactive_debugger(self.args.debugger)
            self.args.interactive_debugger = True
        if self.args.use_valgrind:
            if self.args.env.endswith('existing-env'):
                print(Colors.Bred('can not use valgrind with existing-env'))
                sys.exit(1)
            if self.args.vg_options is None:
                self.args.vg_options = os.getenv('VG_OPTIONS', '--leak-check=full --errors-for-leak-kinds=definite')
            vg_debugger = debuggers.Valgrind(options=self.args.vg_options,
                                             suppressions=self.args.vg_suppressions,
                                             fail_on_errors=not(self.args.vg_no_fail_on_errors),
                                             leakcheck=not(self.args.vg_no_leakcheck)
            )
            if self.args.vg_no_leakcheck:
                vg_debugger.leakcheck = False
            if self.args.no_output_catch or self.args.vg_verbose:
                vg_debugger.verbose = True
            debugger = vg_debugger
        elif self.args.interactive_debugger:
            debugger = debuggers.default_interactive_debugger

        sanitizer = None
        if self.args.sanitizer:
            sanitizer = self.args.sanitizer

        if self.args.env.endswith('existing-env'):
            # when running on existing env we always reuse it
            self.args.env_reuse = True

        # unless None, they must match in length
        if self.args.module_args:
            len_module_args = len(self.args.module_args)
            modules = self.args.module
            if type(modules) == list:
                if (len(modules) != len_module_args):
                    print(Colors.Bred('Using `--module` multiple time implies that you specify the `--module-args` in the the same number'))
                    sys.exit(1)

        Defaults.module = fix_modules(self.args.module)
        Defaults.module_args = fix_modulesArgs(Defaults.module, self.args.module_args)
        Defaults.env = self.args.env
        Defaults.binary = self.args.oss_redis_path
        Defaults.verbose = self.args.verbose
        Defaults.logdir = self.args.log_dir
        Defaults.loglevel = self.args.log_level
        Defaults.use_slaves = self.args.use_slaves
        Defaults.num_shards = self.args.shards_count
        Defaults.shards_ports = self.args.shards_ports.split(',') if self.args.shards_ports is not None else None
        Defaults.cluster_address = self.args.cluster_address
        Defaults.cluster_credentials = self.args.cluster_credentials
        Defaults.internal_password = self.args.internal_password
        Defaults.proxy_binary = self.args.proxy_binary_path
        Defaults.re_binary = self.args.enterprise_redis_path
        Defaults.re_libdir = self.args.enterprise_lib_path
        Defaults.use_aof = self.args.use_aof
        Defaults.debug_pause = self.args.debug
        Defaults.debug_print = self.args.debug_print
        Defaults.no_capture_output = self.args.no_output_catch
        Defaults.print_verbose_information_on_failure = self.args.verbose_information_on_failure
        Defaults.debugger = debugger
        Defaults.sanitizer = sanitizer
        Defaults.exit_on_failure = self.args.exit_on_failure
        Defaults.port = self.args.redis_port
        Defaults.external_addr = self.args.existing_env_addr
        Defaults.use_unix = self.args.unix
        Defaults.randomize_ports = self.args.randomize_ports
        Defaults.use_TLS = self.args.tls
        Defaults.tls_cert_file = self.args.tls_cert_file
        Defaults.tls_key_file = self.args.tls_key_file
        Defaults.tls_ca_cert_file = self.args.tls_ca_cert_file
        Defaults.tls_passphrase = self.args.tls_passphrase
        Defaults.oss_password = self.args.oss_password
        Defaults.cluster_node_timeout = self.args.cluster_node_timeout
        Defaults.enable_debug_command = self.args.enable_debug_command
        Defaults.enable_protected_configs = self.args.enable_protected_configs
        if Defaults.use_unix and Defaults.use_slaves:
            raise Exception('Cannot use unix sockets with slaves')

        self.tests = []
        self.testsFailed = []
        self.currEnv = None
        self.loader = TestLoader()
        if self.args.test is not None:
            self.loader.load_spec(self.args.test)
        if self.args.tests_file is not None:
            for fname in self.args.tests_file:
                try:
                    with open(fname, 'r') as file:
                        for line in file.readlines():
                            line = line.strip()
                            if line.startswith('#') or line == "":
                                continue
                            try:
                                self.loader.load_spec(line)
                            except:
                                print(Colors.Red('Invalid test {TEST} in file {FILE}'.format(TEST=line, FILE=fname)))
                except:
                    print(Colors.Red('Test file {} not found'.format(fname)))
        if self.args.test is None and self.args.tests_file is None:
            self.loader.scan_dir(os.getcwd())

        if self.args.collect_only:
            self.loader.print_tests()
            sys.exit(0)
        if self.args.use_valgrind or self.args.check_exitcode:
            self.require_clean_exit = True
        else:
            self.require_clean_exit = False

        self.parallelism = self.args.parallelism

    def _convertArgsType(self):
        pass

    def takeEnvDown(self, fullShutDown=False):
        if not self.currEnv:
            return

        needShutdown = True
        if self.args.env_reuse and not fullShutDown:
            try:
                self.currEnv.flush()
                needShutdown = False
            except Exception as e:
                self.currEnv.stop()
                self.handleFailure(exception=e, testname='[env dtor]',
                                   env=self.currEnv)

        if needShutdown:
            if self.currEnv.isUp():
                try:
                    self.currEnv.flush()
                    flush_ok = True
                except:
                    flush_ok = False
            self.currEnv.stop()
            if self.require_clean_exit and self.currEnv and (not self.currEnv.checkExitCode() or not flush_ok):
                print(Colors.Bred('\tRedis did not exit cleanly'))
                self.addFailure(self.currEnv.testName, ['redis process failure'])
                if self.args.check_exitcode:
                    raise Exception('Process exited dirty')
            self.currEnv = None

    def printException(self, err):
        msg = 'Unhandled exception: {}'.format(err)
        print('\t' + Colors.Bred(msg))
        traceback.print_exc(file=sys.stdout)

    def addFailuresFromEnv(self, name, env):
        """
        Extract the list of failures from the given test Env
        :param name: The name of the test that failed
        :param env: The Environment which contains the failures
        """
        if not env:
            self.addFailure(name, ['<unknown (environment destroyed)>'])
        else:
            self.addFailure(name, failures=env.assertionFailedSummary)

    def addFailure(self, name, failures=None):
        """
        Adds a list of failures to the report
        :param name: The name of the test that has failures
        :param failures: A string or of strings describing the individual failures
        """
        if failures and not isinstance(failures, (list, tuple)):
            failures = [failures]
        if not failures:
            failures = []
        self.testsFailed.append([name, failures])

    def getTotalFailureCount(self):
        ret = 0
        for _, failures in self.testsFailed:
            ret += len(failures)
        return ret

    def handleFailure(self, testFullName=None, exception=None, prefix='', testname=None, env=None):
        """
        Failure omni-function.

        This function handles failures given a set of input parameters.
        At least one of these must not be empty
        :param exception: The exception to report, of any
        :param prefix: The prefix to use for logging.
            This is usually the test name
        :param testname: The test name, use for recording the failures
        :param env: The environment, used for extracting failed assertions
        """
        if not testname and env:
            testname = env.testName
        elif not testname:
            if prefix:
                testname = prefix
            else:
                testname = '<unknown>'

        if exception:
            self.printError(testFullName if testFullName is not None else '')
            self.printException(exception)
        else:
            self.printFail(testFullName if testFullName is not None else '')

        if env:
            self.addFailuresFromEnv(testname, env)
        elif exception:
            self.addFailure(testname, str(exception))
        else:
            self.addFailure(testname, '<No exception or environment>')

    def _runTest(self, test, numberOfAssertionFailed=0, prefix='', before=None, after=None):
        test.initialize()

        msgPrefix = test.name

        testFullName = prefix + test.name

        if not test.is_method:
            Defaults.curr_test_name = testFullName

        try:
            # Python < 3.11
            test_args = inspect.getargspec(test.target).args
        except:
            test_args = inspect.getfullargspec(test.target).args

        if len(test_args) > 0 and not test.is_method:
            try:
                # env = Env(testName=test.name)
                env = Defaults.env_factory(testName=test.name)
            except Exception as e:
                self.handleFailure(testFullName=testFullName, exception=e, prefix=msgPrefix, testname=test.name)
                return 0

            fn = lambda: test.target(env)
            before_func = (lambda: before(env)) if before is not None else None
            after_func = (lambda: after(env)) if after is not None else None
        else:
            fn = test.target
            before_func = before
            after_func = after

        hasException = False
        try:
            if before_func:
                before_func()
            fn()
            passed = True
        except unittest.SkipTest:
            self.printSkip(testFullName)
            return 0
        except TestAssertionFailure:
            if self.args.exit_on_failure:
                self.takeEnvDown(fullShutDown=True)

            # Don't fall-through
            raise
        except Exception as err:
            if self.args.exit_on_failure:
                self.takeEnvDown(fullShutDown=True)
                after = None
                raise

            self.handleFailure(testFullName=testFullName, exception=err, prefix=msgPrefix,
                               testname=test.name, env=self.currEnv)
            hasException = True
            passed = False
        finally:
            if after_func:
                after_func()

        numFailed = 0
        if self.currEnv:
            numFailed = self.currEnv.getNumberOfFailedAssertion()
            if numFailed > numberOfAssertionFailed:
                self.handleFailure(testFullName=testFullName, prefix=msgPrefix,
                                   testname=test.name, env=self.currEnv)
                passed = False
        elif not hasException:
            self.addFailure(test.name, '<Environment destroyed>')
            passed = False

        # Handle debugger, if needed
        if self.args.stop_on_failure and not passed:
            if self.args.interactive_debugger:
                while self.currEnv.isUp():
                    time.sleep(1)
            input('press any button to move to the next test')

        if passed:
            self.printPass(testFullName)

        if hasException:
            numFailed += 1 # exception should be counted as failure
        return numFailed

    def printSkip(self, name):
        print('%s:\r\n\t%s' % (Colors.Cyan(name), Colors.Green('[SKIP]')))

    def printFail(self, name):
        print('%s:\r\n\t%s' % (Colors.Cyan(name), Colors.Bred('[FAIL]')))

    def printError(self, name):
        print('%s:\r\n\t%s' % (Colors.Cyan(name), Colors.Bred('[ERROR]')))

    def printPass(self, name):
        print('%s:\r\n\t%s' % (Colors.Cyan(name), Colors.Green('[PASS]')))

    def envScopeGuard(self):
        return EnvScopeGuard(self)

    def execute(self):
        Env.RTestInstance = self
        if self.args.env_only:
            Defaults.verbose = 2
            # env = Env(testName='manual test env')
            env = Defaults.env_factory(testName='manual test env')
            if self.args.interactive_debugger:
                while env.isUp():
                    time.sleep(1)
            else:
                cmd = MyCmd(env)
                cmd.cmdloop()
            env.stop()
            return
        done = 0
        startTime = time.time()
        if self.args.interactive_debugger and len(self.loader.tests) != 1:
            print(self.tests)
            print(Colors.Bred('only one test can be run on interactive-debugger use -t'))
            sys.exit(1)

        jobs = Queue()
        for test in self.loader:
            jobs.put(test, block=False)

        def run_jobs(jobs, results, port):
            Defaults.port = port
            done = 0
            while True:
                try:
                    test = jobs.get(timeout=0.1)
                except Exception as e:
                    break

                with self.envScopeGuard():
                    if test.is_class:
                        test.initialize()

                        Defaults.curr_test_name = test.name
                        try:
                            obj = test.create_instance()

                        except unittest.SkipTest:
                            self.printSkip(test.name)
                            continue

                        except Exception as e:
                            self.printException(e)
                            self.addFailure(test.name + " [__init__]")
                            continue

                        failures = 0
                        before = getattr(obj, 'setUp', None)
                        after = getattr(obj, 'tearDown', None)
                        for subtest in test.get_functions(obj):
                            failures += self._runTest(subtest, prefix='\t',
                                                    numberOfAssertionFailed=failures,
                                                    before=before, after=after)
                            done += 1

                    else:
                        failures = self._runTest(test)
                        done += 1

                    verboseInfo = {}
                    if failures > 0 and Defaults.print_verbose_information_on_failure:
                        lastEnv = self.currEnv
                        verboseInfo['before_dispose'] = lastEnv.getInformationBeforeDispose()

                # here the env is down so lets collect more info and print it
                if failures > 0 and Defaults.print_verbose_information_on_failure:
                    verboseInfo['after_dispose'] = lastEnv.getInformationAfterDispose()
                    lastEnv.debugPrint(json.dumps(verboseInfo, indent=2).replace('\\n', '\n'), force=True)

            self.takeEnvDown(fullShutDown=True)

            # serialized the results back
            results.put({'done': done, 'failures': self.testsFailed}, block=False)

        results = Queue()
        if self.parallelism == 1:
            run_jobs(jobs, results, Defaults.port)
        else :
            processes = []
            currPort = Defaults.port
            for i in range(self.parallelism):
                p = Process(target=run_jobs, args=(jobs,results,currPort))
                currPort += 30 # safe distance for cluster and replicas
                processes.append(p)
                p.start()

            for p in processes:
                p.join()

        # join results
        while True:
            try:
                res = results.get(timeout=0.1)
            except Exception as e:
                break
            done += res['done']
            self.testsFailed.extend(res['failures'])

        endTime = time.time()

        print(Colors.Bold('Test Took: %d sec' % (endTime - startTime)))
        print(Colors.Bold('Total Tests Run: %d, Total Tests Failed: %d, Total Tests Passed: %d' % (done, self.getTotalFailureCount(), done - self.getTotalFailureCount())))
        if self.testsFailed:
            if self.args.failed_tests_file:
                with open(self.args.failed_tests_file, 'w') as file:
                    for test, _ in self.testsFailed:
                        file.write(test.split(' ')[0] + "\n")

            print(Colors.Bold('Failed Tests Summary:'))
            for group, failures in self.testsFailed:
                print('\t' + Colors.Bold(group))
                if not failures:
                    print('\t\t' + Colors.Bred('Exception raised during test execution. See logs'))
                for failure in failures:
                    print('\t\t' + failure)
            sys.exit(1)
        else:
            if self.args.failed_tests_file:
                with open(self.args.failed_tests_file, 'w') as file:
                    pass


def main():
    # Aviod "UnicodeEncodeError: 'ascii' codec can't encode character" errors
    sys.stdout = io.open(sys.stdout.fileno(), 'w', encoding='utf8')
    sys.stderr = io.open(sys.stderr.fileno(), 'w', encoding='utf8')
    RLTest().execute()


if __name__ == '__main__':
    main()



from __future__ import print_function

import os
import psutil
import subprocess
from RLTest.utils import Colors


class Dmc():
    DMC_LOG_FILE_NAME = 'dmc.log'

    def __init__(self, dmcBinaryPath, libPath, directory=None):
        self.dmcBinaryPath = os.path.expanduser(dmcBinaryPath) if dmcBinaryPath.startswith('~/') else dmcBinaryPath
        self.env = {'MEMTIER_NODE_ID': '1'}
        self.libPath = libPath
        if 'LD_LIBRARY_PATH' in os.environ:
            self.env['LD_LIBRARY_PATH'] = os.environ['LD_LIBRARY_PATH']
        else:
            self.env['LD_LIBRARY_PATH'] = os.path.expanduser(self.libPath) if dmcBinaryPath.startswith('~/') else self.libPath
        self.directory = directory

        logFile = self.DMC_LOG_FILE_NAME if self.directory is None else os.path.join(self.directory, self.DMC_LOG_FILE_NAME)
        self.proxyArgs = [
            self.dmcBinaryPath,
            '-c', '500',
            '-O', logFile
        ]

    def Start(self):
        self.process = psutil.Popen(executable=self.dmcBinaryPath,
                                    stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                    env=self.env,
                                    args=self.proxyArgs)

    def Stop(self):
        if self.process:
            self.process.kill()
            self.process.wait()
        self.process = None

    def PrintEnvData(self, prefix=''):
        print(Colors.Yellow(prefix + 'dmc binary path: %s' % self.dmcBinaryPath))
        print(Colors.Yellow(prefix + 'dmc env: %s' % str(self.env)))
        print(Colors.Yellow(prefix + 'dir path: %s' % str(self.directory)))
        print(Colors.Yellow(prefix + 'log file name: %s' % str(self.DMC_LOG_FILE_NAME)))
        if self.libPath:
            print(Colors.Yellow(prefix + 'lib path: %s' % str(self.libPath)))



import random
import socket
import struct
import fcntl
import os
import errno
import json

def _check_alive(pid):
    try:
        os.kill(pid, 0)
        return True
    except OSError as e:
        if e.errno == errno.EPERM:
            return True
        return False


def register_port(port):
    fp = open('/tmp/rltest_portfile.lock', 'a+')
    fcntl.flock(fp.fileno(), fcntl.LOCK_EX)
    fp.seek(0, 2)  # seek from end
    if fp.tell() == 0:
        entries = {}
    else:
        fp.seek(0, 0)
        entries = json.load(fp)
    # remove not responsive processes
    entries = { p:pid for p, pid in entries.items() if _check_alive(pid) }

    if str(port) in entries:
        ret = False
    else:
        entries[str(port)] = os.getpid()
        ret = True

    fp.seek(0, 0)
    fp.truncate()
    json.dump(entries, fp)
    fp.close()
    return ret


def get_random_port():
    for _ in range(10000):
        p = random.randint(10000, 20000)
        # Try to open and bind the socket
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, struct.pack('ii', 1, 0))
        try:
            s.bind(('', p))
            s.close()
            if not register_port(p):
                continue
            return p
        except Exception as e:
            if hasattr(e, 'errno') and e.errno in (errno.EADDRINUSE, errno.EADDRNOTAVAIL):
                pass
            else:
                raise e

    raise Exception('Could not find open port to listen on!')



from __future__ import print_function
import os
import sys
import imp
import inspect
from RLTest.utils import Colors


class TestFunction(object):
    is_class = False

    def __init__(self, filename, symbol, modulename):
        self.filename = filename
        self.symbol = symbol
        self.modulename = modulename
        self.is_method = False
        self.name = '{}:{}'.format(self.modulename, symbol)

    def initialize(self):
        module_file = open(self.filename)
        module = imp.load_module(self.modulename, module_file, self.filename, ('.py', 'r', imp.PY_SOURCE))
        obj = getattr(module, self.symbol)
        self.target = obj

    def shortname(self):
        return self.target.__name__

class TestMethod(object):
    is_class = False

    def __init__(self, obj, name):
        self.target = obj
        self.name = name
        self.is_method = True

    def initialize(self):
        pass

    def shortname(self):
        return self.target.__name__

class TestClass(object):
    is_class = True

    def __init__(self, filename, symbol, modulename, functions):
        self.filename = filename
        self.symbol = symbol
        self.modulename = modulename
        self.functions = functions
        self.name = '{}:{}'.format(self.modulename, symbol)

    def initialize(self):
        module_file = open(self.filename)
        module = imp.load_module(self.modulename, module_file, self.filename, ('.py', 'r', imp.PY_SOURCE))
        obj = getattr(module, self.symbol)
        self.clsname = obj.__name__
        self.cls = obj

    def create_instance(self, *args, **kwargs):
        return self.cls(*args, **kwargs)

    def get_functions(self, instance):
        fns = []
        for mname in self.functions:
            bound = getattr(instance, mname)
            if not callable(bound):
                continue
            fns.append(TestMethod(bound,
                                  name='{}:{}.{}'.format(self.modulename, self.clsname, mname)))
        return fns


class TestLoader(object):
    def __init__(self):
        self.tests = []

    def load_spec(self, arg):
        # if arg is a list, load its elements
        if isinstance(arg, list):
            for spec in arg:
                self.load_spec(spec)
            return

        # See what kind of spec this is!
        """
        Load tests from single argument form, e.g. foo.py:BarBaz
        """
        if ':' in arg:
            filename, varname = arg.split(':')
        else:
            filename = arg
            varname = None
            if os.path.isdir(filename):
                sys.path.append(filename)
                self.scan_dir(filename)
                return

        # Ensure the path is in sys.path
        dirname = os.path.abspath(os.path.dirname(filename))
        if dirname not in sys.path:
            sys.path.append(dirname)

        module_name, _ = os.path.splitext(os.path.basename(filename))
        toplevel_filter, subfilter = None, None
        if varname:
            if '.' in varname:
                toplevel_filter, subfilter = varname.split('.')
            else:
                toplevel_filter = varname

        self.load_files(dirname, module_name, toplevel_filter, subfilter)

    def load_files(self, module_dir, module_name, toplevel_filter=None, subfilter=None):
        filename = '%s/%s.py' % (module_dir, module_name)
        try:
            with open(filename, 'r') as module_file:
                try:
                    module = imp.load_module(module_name, module_file, filename,
                                             ('.py', 'r', imp.PY_SOURCE))
                    for symbol in dir(module):
                        if not self.filter_modulevar(symbol, toplevel_filter):
                            continue

                        obj = getattr(module, symbol)
                        if inspect.isclass(obj):
                            methnames = [mname for mname in dir(obj)
                                         if self.filter_method(mname, subfilter)]
                            self.tests.append(TestClass(filename, symbol, module_name, methnames))
                        elif inspect.isfunction(obj):
                            self.tests.append(TestFunction(filename, symbol, module_name))
                except Exception as x:
                    print(Colors.Red("Problems in file %s: %s" % (filename, x)))
        except:
            print(Colors.Red("File %s not found: skipping" % filename))

    def scan_dir(self, testdir):
        for filename in os.listdir(testdir):
            if filename.startswith('test') and filename.endswith('.py'):
                module_name, ext = os.path.splitext(filename)
                self.load_files(testdir, module_name)

    def filter_modulevar(self, candidate, toplevel_filter):
        if not candidate.lower().startswith('test'):
            return False
        if toplevel_filter and candidate != toplevel_filter:
            return False
        return True

    def filter_method(self, candidate, subfilter):
        if not candidate.lower().startswith('test'):
            return False
        if subfilter and candidate != subfilter:
            return False

        return True

    def __iter__(self):
        return iter(self.tests)

    def print_tests(self):
        tests = []
        for t in self.tests:
            if t.is_class:
                for m in t.functions:
                    tests.append(f"{t.name}.{m}")
            else:
                tests.append(t.name)
        print(*sorted(tests), sep='\n')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['RLTest', 'RLTest.Enterprise']

package_data = \
{'': ['*']}

install_requires = \
['distro>=1.5.0,<2.0.0',
 'psutil>=5.9.5,<6.0.0',
 'pytest-cov>=4.1.0,<5.0.0',
 'pytest>=7.4,<8.0',
 'redis>=5.0.0rc2,<6.0.0']

entry_points = \
{'console_scripts': ['RLTest = RLTest.__main__:main']}

setup_kwargs = {
    'name': 'rltest',
    'version': '0.7.2',
    'description': 'Redis Modules Test Framework, allow to run tests on redis and modules on a variety of environments',
    'long_description': '[![license](https://img.shields.io/github/license/RedisLabsModules/RLTest.svg)](https://github.com/RedisLabsModules/RLTest/blob/master/LICENSE)\n[![PyPI version](https://badge.fury.io/py/rltest.svg)](https://badge.fury.io/py/rltest)\n[![CI](https://github.com/RedisLabsModules/RLTest/workflows/CI/badge.svg)](https://github.com/RedisLabsModules/RLTest/actions)\n[![Version](https://img.shields.io/github/release/RedisLabsModules/RLTest.svg)](https://github.com/RedisLabsModules/RLTest/releases/latest)\n[![Codecov](https://codecov.io/gh/RedisLabsModules/RLTest/branch/master/graph/badge.svg)](https://codecov.io/gh/RedisLabsModules/RLTest)\n[![Known Vulnerabilities](https://snyk.io/test/github/RedisLabsModules/RLTest/badge.svg?targetFile=pyproject.toml)](https://snyk.io/test/github/RedisLabsModules/RLTest?targetFile=pyproject.toml)\n\n\n\n# RLTest\n\nRedis Labs Test Framework, allow running tests on redis and modules on a variety of environments.\n\nSupported Environment: oss, oss-cluster, enterprise, enterprise-cluster\n\nThe framework allow you to write a test without environment specification and then run the test on all supported environment.\n\n\n# Install\n```\n$ pip install git+https://github.com/RedisLabsModules/RLTest.git@master\n\n```\n\n# Usage:\n```\n$ RLTest --help\nusage: RLTest [-h] [--version] [--module MODULE] [--module-args MODULE_ARGS]\n              [--env {oss,oss-cluster,enterprise,enterprise-cluster,existing-env,cluster_existing-env}]\n              [--existing-env-addr EXISTING_ENV_ADDR]\n              [--shards_ports SHARDS_PORTS]\n              [--cluster_address CLUSTER_ADDRESS]\n              [--oss_password OSS_PASSWORD]\n              [--cluster_credentials CLUSTER_CREDENTIALS]\n              [--cluster_node_timeout CLUSTER_NODE_TIMEOUT]\n              [--internal_password INTERNAL_PASSWORD]\n              [--oss-redis-path OSS_REDIS_PATH]\n              [--enterprise-redis-path ENTERPRISE_REDIS_PATH]\n              [--stop-on-failure] [-x] [--verbose] [--debug] [-t TEST]\n              [--env-only] [--clear-logs] [--log-dir LOG_DIR] [--use-slaves]\n              [--shards-count SHARDS_COUNT] [--download-enterprise-binaries]\n              [--proxy-binary-path PROXY_BINARY_PATH]\n              [--enterprise-lib-path ENTERPRISE_LIB_PATH] [-r]\n              [--use-aof] [--use-rdb-preamble]\n              [--debug-print] [-V] [--vg-suppressions VG_SUPPRESSIONS]\n              [--vg-options VG_OPTIONS] [--vg-no-leakcheck] [--vg-verbose]\n              [--vg-no-fail-on-errors] [-i] [--debugger DEBUGGER] [-s]\n              [--check-exitcode] [--unix] [--randomize-ports] [--collect-only]\n              [--tls] [--tls-cert-file TLS_CERT_FILE]\n              [--tls-key-file TLS_KEY_FILE]\n              [--tls-ca-cert-file TLS_CA_CERT_FILE]\n\nTest Framework for redis and redis module\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --version             Print RLTest version and exit (default: False)\n  --module MODULE       path to the module file. You can use `--module` more\n                        than once but it imples that you explicitly specify\n                        `--module-args` as well. Notice that on enterprise the\n                        file should be a zip file packed with\n                        [RAMP](https://github.com/RedisLabs/RAMP). (default:\n                        None)\n  --module-args MODULE_ARGS\n                        arguments to give to the module on loading (default:\n                        None)\n  --env {oss,oss-cluster,enterprise,enterprise-cluster,existing-env,cluster_existing-env}, -e {oss,oss-cluster,enterprise,enterprise-cluster,existing-env,cluster_existing-env}\n                        env on which to run the test (default: oss)\n  --existing-env-addr EXISTING_ENV_ADDR\n                        Address of existing env, relevent only when running\n                        with existing-env, cluster_existing-env (default:\n                        localhost:6379)\n  --shards_ports SHARDS_PORTS\n                        list of ports, the shards are listening to, relevent\n                        only when running with cluster_existing-env (default:\n                        None)\n  --cluster_address CLUSTER_ADDRESS\n                        enterprise cluster ip, relevent only when running with\n                        cluster_existing-env (default: None)\n  --oss_password OSS_PASSWORD\n                        set redis password, relevant for oss and oss-cluster\n                        environment (default: None)\n  --cluster_credentials CLUSTER_CREDENTIALS\n                        enterprise cluster cluster_credentials\n                        "username:password", relevent only when running with\n                        cluster_existing-env (default: None)\n  --cluster_node_timeout CLUSTER_NODE_TIMEOUT\n                        cluster node timeout in milliseconds\n  --internal_password INTERNAL_PASSWORD\n                        Give an ability to execute commands on shards\n                        directly, relevent only when running with\n                        cluster_existing-env (default: )\n  --oss-redis-path OSS_REDIS_PATH\n                        path to the oss redis binary (default: redis-server)\n  --enterprise-redis-path ENTERPRISE_REDIS_PATH\n                        path to the entrprise redis binary (default:\n                        ~/.RLTest/opt/redislabs/bin/redis-server)\n  --stop-on-failure     stop running on failure (default: False)\n  -x, --exit-on-failure\n                        Stop test execution and exit on first assertion\n                        failure (default: False)\n  --verbose, -v         print more information about the test (default: 0)\n  --debug               stop before each test allow gdb attachment (default:\n                        False)\n  -t TEST, --test TEST  Specify test to run, in the form of "file:test"\n                        (default: None)\n  --env-only            start the env but do not run any tests (default:\n                        False)\n  --clear-logs          deleting the log direcotry before the execution\n                        (default: False)\n  --log-dir LOG_DIR     directory to write logs to (default: ./logs)\n  --use-slaves          run env with slaves enabled (default: False)\n  --shards-count SHARDS_COUNT\n                        Number shards in bdb (default: 1)\n  --download-enterprise-binaries\n                        run env with slaves enabled (default: False)\n  --proxy-binary-path PROXY_BINARY_PATH\n                        dmc proxy binary path (default:\n                        ~/.RLTest/opt/redislabs/bin/dmcproxy)\n  --enterprise-lib-path ENTERPRISE_LIB_PATH\n                        path of needed libraries to run enterprise binaries\n                        (default: ~/.RLTest/opt/redislabs/lib/)\n  -r, --env-reuse       reuse exists env, this feature is based on best\n                        efforts, if the env can not be reused then it will be\n                        taken down. (default: False)\n  --use-aof             use aof instead of rdb (default: False)\n  --use-rdb-preamble    use rdb preamble when rewriting aof file (default: True)\n  --debug-print         print debug messages (default: False)\n  -V, --vg, --use-valgrind\n                        running redis under valgrind (assuming valgrind is\n                        install on the machine) (default: False)\n  --vg-suppressions VG_SUPPRESSIONS\n                        path valgrind suppressions file (default: None)\n  --vg-options VG_OPTIONS\n                        valgrind [options] (default: None)\n  --vg-no-leakcheck     Don\'t perform a leak check (default: False)\n  --vg-verbose          Don\'t log valgrind output. Output to screen directly\n                        (default: False)\n  --vg-no-fail-on-errors\n                        Dont Fail test when valgrind reported any errors in\n                        the run.By default on RLTest the return value from\n                        Valgrind will be used to fail the tests.Use this\n                        option when you wish to dry-run valgrind but not fail\n                        the test on valgrind reported errors. (default: False)\n  -i, --interactive-debugger\n                        runs the redis on a debuger (gdb/lldb)\n                        interactivly.debugger interactive mode is only\n                        possible on a single process and so unsupported on\n                        cluste or with slaves.it is also not possible to use\n                        valgrind on interactive mode.interactive mode direcly\n                        applies: --no-output-catch and --stop-on-failure.it is\n                        also implies that only one test will be run (if --env-\n                        only was not specify), an error will be raise\n                        otherwise. (default: False)\n  --debugger DEBUGGER   Run specified command line as the debugger (default:\n                        None)\n  -s, --no-output-catch\n                        all output will be written to the stdout, no log\n                        files. (default: False)\n  --check-exitcode      Check redis process exit code (default: False)\n  --unix                Use Unix domain sockets instead of TCP (default:\n                        False)\n  --randomize-ports     Randomize Redis listening port assignment rather\n                        thanusing default port (default: False)\n  --collect-only        Collect the tests and exit (default: False)\n  --tls                 Enable TLS Support and disable the non-TLS port\n                        completely. TLS connections will be available at the\n                        default non-TLS ports. (default: False)\n  --tls-cert-file TLS_CERT_FILE\n                        /path/to/redis.crt (default: None)\n  --tls-key-file TLS_KEY_FILE\n                        /path/to/redis.key (default: None)\n  --tls-ca-cert-file TLS_CA_CERT_FILE\n                        /path/to/ca.crt (default: None)\n\n```\n\n## Sample usages\n\n### Multiple modules\n\n```\nRLTest --module modules/module1.so --module-args \'\' --module modules/module2.so --module-args \'\'\n```\n\n# Configuration File\nBy default, the framework search for configuration file on the current directory. The configuration file name is: config.txt.\nIt is possible to specify different configuration file on command line using the \'@\' prefix, for example:\n```\nRLTest @myConfig.txt # search for myConfig.txt configuration file\n```\nThe configuration file format is the same as the command line argument, i.e : \'--< param_name > < param_val >\'.\n\nIt is also possible to comment a specific lines in the configuration file using \'#\'.\n\nExample:\n\n```\n-vv\n--clear-logs\n#--debug\n```\n\n\n# Test Example\n\n```python\nfrom RLTest import Env\nimport time\n\n\nclass testExample():\n    \'\'\'\n    run all tests on a single env without taking\n    env down between tests\n    \'\'\'\n    def __init__(self):\n        self.env = Env()\n\n    def testExample(self):\n        con = self.env.getConnection()\n        con.set(\'x\', 1)\n        self.env.assertEqual(con.get(\'x\'), \'1\')\n\n    def testExample1(self):\n        con = self.env.getConnection()\n        con.set(\'x\', 1)\n        self.env.assertEqual(con.get(\'x\'), \'1\')\n        self.env.assertFalse(True)  # check failure\n\n    def testExample2(self):\n        con = self.env.getConnection()\n        con.set(\'x\', 1)\n        self.env.assertEqual(con.get(\'x\'), \'1\')\n\n\n# run each test on different env\ndef test_example():\n    env = Env()\n    con = env.getConnection()\n    con.set(\'x\', 1)\n    env.assertEqual(con.get(\'x\'), \'1\')\n\n\ndef test_example_2():\n    env = Env()\n    env.assertOk(env.cmd(\'set\', \'x\', \'1\'))\n    env.expect(\'get\', \'x\').equal(\'1\')\n\n    env.expect(\'lpush\', \'list\', \'1\', \'2\', \'3\').equal(3)\n    env.expect(\'lrange\', \'list\', \'0\', \'-1\').debugPrint().contains(\'1\')\n    env.debugPrint(\'this is some debug printing\')\n\n\ndef test_example_3():\n    env = Env(useSlaves=True, env=\'oss\')\n    con = env.getConnection()\n    con.set(\'x\', 1)\n    con2 = env.getSlaveConnection()\n    time.sleep(0.1)\n    env.assertEqual(con2.get(\'x\'), \'1\')\n\n```\n',
    'author': 'Redis, Inc.',
    'author_email': 'oss@redis.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0',
}


setup(**setup_kwargs)
